import warnings
from typing import List, Tuple, Dict
from Bio.Seq import Seq
from Bio.PDB import PPBuilder
from Bio.Data.PDBData import protein_letters_3to1_extended
from Bio.Data.PDBData import nucleic_letters_3to1_extended
from Bio.PDB.Chain import Chain as _Chain
from Bio.PDB.PDBExceptions import PDBConstructionException
from crimm.StructEntities.Residue import DisorderedResidue

class BaseChain(_Chain):
    """Base class derived from and Biopython chain object and compatible with
    Biopython's functions"""
    chain_type = "Base Chain"
    _res_repr_str = "Residues"
    def __init__(self, chain_id):
        super().__init__(chain_id)
        # the following attributes are used for topology definitions
        # instantiated but not initiated
        self.undefined_res = None
        self.topo_definitions = None
        self.topology = None
        self.pdbx_description = None

    @property
    def segids(self):
        """Return the list of segids in this chain."""
        segid_set = set()
        for res in self:
            if res.segid is not None and (res.segid not in ('', ' ')):
                segid_set.add(res.segid)
        return list(segid_set)
    
    @property
    def residues(self):
        """Alias for child_list. Returns the list of residues in this chain."""
        return self.child_list

    @property
    def total_charge(self):
        """Return the total charge of the chain."""
        total_charge = 0
        for res in self:
            if res.total_charge is None:
                return None
            total_charge += res.total_charge
        return round(total_charge, 2)
    
    def sort_residues(self):
        """Update the ordering of the residues in child list by resseq
        """
        self.child_list = sorted(self.child_list, key=lambda x: x.id[1:])
        self.child_dict = {res.id: res for res in self.child_list}

    def truncate(self, start=None, end=None):
        """Truncate the chain to the specified start and end residue id in the 
        residue list (0-indexed). If start is None, the first residue will be used.
        If end is None, the last residue will be used."""
        if start is None:
            start = 0
        if end is None:
            end = len(self)
        self.sort_residues()
        remove_list = []
        remove_list.extend(self.child_list[:start])
        remove_list.extend(self.child_list[end:])
        for res in remove_list:
            res.detach_parent()
            self.detach_child(res.id)
            if hasattr(self, 'het_resseq_lookup'):
                if res.id[1] in self.het_resseq_lookup:
                    self.het_resseq_lookup.pop(res.id[1])
        self.sort_residues()
        

    def truncate_by_seqid(self, start=None, end=None):
        """Truncate the chain to the specified start and end residue sequence number."""
        if start is None:
            start = self.residues[0].id[1]
        if end is None:
            end = self.residues[-1].id[1]
        start = self._translate_id(start)
        end = self._translate_id(end)
        if start not in self.child_dict:
            raise ValueError(f"Start residue {start} not found in the chain.")
        if end not in self.child_dict:
            raise ValueError(f"End residue {end} not found in the chain.")
        start = self.child_dict[start].id
        end = self.child_dict[end].id
        if start > end:
            raise ValueError(f"Start residue {start} is greater than end residue {end}.")
        remove_list = []
        for res in self.child_list:
            if res.id < start or res.id > end:
                remove_list.append(res)
        for res in remove_list:
            res.detach_parent()
            self.detach_child(res.id)
            if hasattr(self, 'het_resseq_lookup'):
                if res.id[1] in self.het_resseq_lookup:
                    self.het_resseq_lookup.pop(res.id[1])
        self.sort_residues()

    def get_top_parent(self):
        if self.parent is None:
            return self
        return self.parent.get_top_parent()

    def reset_atom_serial_numbers(
            self, include_alt = True, reset_current_only = False
        ):
        """Reset all atom serial numbers in the encompassing entity (the parent
        structure and/or model, if they exist) starting from 1. If include_alt
        is True, the disordered atoms will be expanded and the altloc will be
        included in the serial number. If reset_current_only is True, the serial 
        numbers will be reset for the entity itself, otherwise, the top parent 
        will be searched for the reset."""
        top_parent = self.get_top_parent()
        if top_parent is not self and not reset_current_only:
            top_parent.reset_atom_serial_numbers(include_alt=include_alt)
            return
        # no parent, reset the serial number for the entity itself
        i = 1
        for atom in self.get_atoms(include_alt=include_alt):
            atom.set_serial_number(i)
            i+=1

    ## TODO: move this to the structure builder
    def _disordered_reset_residue(self, disordered_residue):
        for resname, child_residue in disordered_residue.child_dict.items():
            if child_residue.id == disordered_residue.id:
                disordered_residue.disordered_select(resname)

    def _disordered_reset_atom(self, residue):
        for atom in residue.get_unpacked_list():
            if atom.is_disordered() == 2:
                first_alt = sorted(atom.child_dict)[0]
                atom.disordered_select(first_alt)

    def reset_disordered_residues(self):
        """Reset the selected child of all disordered residues to the first
        residue (alt loc A) supplied by PDB."""
        for res in self:
            if isinstance(res, DisorderedResidue):
                self._disordered_reset_residue(res)
            elif res.disordered == 1:
                self._disordered_reset_atom(res)

    def __repr__(self):
        """Return the chain identifier."""
        segids = ', '.join(self.segids)
        if len(segids) == 0:
            return f"<{self.chain_type} id={self.get_id()} {self._res_repr_str}={len(self)}>"
        return f"<{self.chain_type} id={self.get_id()} {self._res_repr_str}={len(self)} segids={segids}>"
        
    def expanded_view(self):
        """Print the expanded view of the chain."""
        repr_str = repr(self)
        if (resnames := getattr(self, 'resnames', None)) is not None:
            repr_str += f"\n  Residue ID(s): {resnames}"
        if (descr := getattr(self, 'pdbx_description', None)) is not None:
            repr_str += f"\n  Description: {descr}"
        return repr_str
    
    def _ipython_display_(self):
        """Return the nglview interactive visualization window"""
        if len(self) != 0:
            from crimm.Visualization import show_nglview
            from IPython.display import display
            display(show_nglview(self))

        print(self.expanded_view())

    def get_atoms(self, include_alt = False):
        """Return a generator of all atoms from this chain. If include_alt is True, the 
        disordered residues will be expanded and altloc of disordered atoms will be included."""
        if include_alt:
            all_res = self.get_unpacked_list()
        else:
            all_res = self.child_list
        for res in all_res:
            yield from res.get_atoms(include_alt=include_alt)

    def is_continuous(self):
        """Not implemented in BaseChain. Implementation varies depending on the
        child class"""
        raise NotImplementedError

class Chain(BaseChain):
    """
    General/Unspecified Chain object based on Biopython Chain
    """
    chain_type = 'Chain'
    ## TODO: Implement hetflag check for florescence proteins (chromophore residues)
    def __init__(self, chain_id: str):
        super().__init__(chain_id)
        self._ppb = PPBuilder()
        # There is no way to distinguish the type of the chain. Hence, the
        # nucleic letter codes are limited to one letter codes
        self.letter_3to1_dict = {
            **protein_letters_3to1_extended,
            **{v:v for k, v in nucleic_letters_3to1_extended.items()}
        }
        self.het_resseq_lookup = {}

    def add(self, residue):
        """Add a child to the Entity. Overwrite the Biopython Chain.add method"""
        entity_id = residue.get_id()
        hetflag, resseq, icode = entity_id
        if self.has_id(entity_id):
            raise PDBConstructionException(f"{entity_id} defined twice")
        residue.set_parent(self)
        self.child_list.append(residue)
        self.child_dict[entity_id] = residue
        if hetflag.startswith('H_'):
            self.het_resseq_lookup[resseq] = entity_id
        if icode != ' ':
            self.het_resseq_lookup[resseq] = entity_id

    def _translate_id(self, id):
        """Translate sequence identifier to tuple form (PRIVATE).

        A residue id is normally a tuple (hetero flag, sequence identifier,
        insertion code). The _translate_id method first looks up if there is a 
        heterogen residue with the resseq id, if not, it translates the 
        sequence identifier to the (" ", sequence identifier, " ") tuple.

        Arguments:
         - id - int, residue resseq

        """
        if id in self.het_resseq_lookup:
            return self.het_resseq_lookup[id]
        if isinstance(id, int):
            id = (" ", id, " ")
        return id
    
    def get_disordered_res(self):
        disordered_res = []
        for res in self:
            if isinstance(res, DisorderedResidue):
                disordered_res.append(res)
        return disordered_res
    
    def get_segments(self):
        """
        Build polypeptide segments based on C-N distance criterion
        """
        # This will detect the gap in chain better than using residue sequence
        # numbering
        return self._ppb.build_peptides(self, aa_only=False)

    def extract_segment_seq(self):
        """
        Extract sequence from the residues that are present in chain. Residues that 
        have missing C and N backbone atoms are skipped.
        """
        seq = Seq('')
        for segment in self.get_segments():
            seq += segment.get_sequence()
        return seq
    
    def extract_present_seq(self):
        """
        Extract sequence from the residues that are present in chain, regardless
        of any missing atom or even empty residues.
        """
        seq = Seq('')
        for res in sorted(self, key = lambda x: x.id[1:]):
            if len(res.child_list) == 0:
                continue
            if res.resname not in self.letter_3to1_dict:
                one_letter_code = 'X'
            else:
                one_letter_code = self.letter_3to1_dict[res.resname]
            seq+=one_letter_code
        return seq
    
    def is_continuous(self):
        """
        Check if the chain is continuous (no gaps). Missing segments on 
        the C and N terminals will be ignored.
        """
        segments = self.get_segments()
        # If the strucutures C-N distance are all within the criterion,
        # it is continuous
        return len(segments) == 1

class PolymerChain(Chain):
    """
    A derived Chain class for holding only polypeptide or nucleotide residues with 
    sequence id reset to be continuous and start from 1 (MMCIF standard). 
    Author reported canonical sequence and missing residues are required for its
    construction.
    This is the default class for polypeptide and polynucleotide parsed from mmCIF.
    """
    def __init__(
        self,
        chain_id: str,
        entity_id: int,
        author_chain_id: str,
        chain_type: str,
        known_sequence: str,
        canon_sequence: str,
        reported_res: List[Tuple[int, str]],
        reported_missing_res: List[Tuple[int, str]] = None
    ):
        """
        Attributes:
            instance_attribute (str): The instance attribute
        """
        if chain_id is None:
            chain_id = entity_id
        super().__init__(chain_id)
        self.entity_id = entity_id
        self.author_chain_id = author_chain_id
        self.chain_type = chain_type[0].upper()+chain_type[1:]
        self.reported_res = reported_res
        if reported_missing_res is None:
            reported_missing_res = []
        self.reported_missing_res = reported_missing_res
        self.known_seq = Seq(known_sequence)
        self.can_seq = Seq(canon_sequence)
        self.seq_lookup: Dict[int, Tuple[str, int, str]] = {}
        if len(self.reported_res) != len(self.can_seq):
            warnings.warn(
                "Total number of reported residues do not match with the "
                "length of reported canonical sequence."
            )
        if self.chain_type == 'Polypeptide(L)':
            self.letter_3to1_dict = protein_letters_3to1_extended
        else:
            self.letter_3to1_dict = {
                **{v:v for k, v in nucleic_letters_3to1_extended.items()},
                **nucleic_letters_3to1_extended
            }

    def __contains__(self, id):
        return super().__contains__(id) or id in self.het_resseq_lookup

    def __getitem__(self, id):
        # Translate the sequence id to the tuple form. We rely on the one-to-one
        # correspendence here
        if id in self.het_resseq_lookup:
            id = self.het_resseq_lookup[id]
        return super().__getitem__(id)
    @property
    def seq(self):
        """
        Get the current sequence from the present residues only
        """
        return self.extract_present_seq()
    
    @property
    def missing_res(self):
        """
        Get the current missing residues in the chain.
        Currently present residues will be compared to the the list of author- 
        reported residues to determine the missing ones.
        """
        present_res = set()
        for res in self:
            if isinstance(res, DisorderedResidue):
                # Record both residues as present from the disordered residue
                for resname, child_res in res.child_dict.items():
                    _, resseq, _ = child_res.get_id()
                    present_res.add((resseq, resname))
            else:
                if res.resname in ('HSP', 'HSE', 'HSD'):
                    # HIS residues are reported as HSP, HSE, HSD in CHARMM force field
                    resname = 'HIS'
                else:
                    resname = res.resname
                _, resseq, _ = res.get_id()
                present_res.add((resseq, resname))



        missing_res = []
        for i, reported in self.reported_res:
            if (i, reported) not in present_res:
                missing_res.append((i, reported))

        return missing_res

    @property
    def masked_seq(self):
        """
        Get a sequence masked with '-' for any residue that is missing CA and N
        backbone atoms
        """
        if len(self.missing_res) == 0:
            missing_res_ids = []
        else:
            missing_res_ids = list(zip(*self.missing_res))[0]
        return MaskedSeq(missing_res_ids, self.can_seq, self.reported_res[0][0])

    @property
    def gaps(self):
        """
        Group gap residues into sets for comparison purposes
        """
        missing_res = self.missing_res
        gaps = []
        if len(missing_res) == 0:
            return gaps
        
        prev_idx = missing_res[0][0]
        cur_set = set()
        for idx, res_name in missing_res[1:]:
            cur_set.add(prev_idx)
            if idx - prev_idx > 1:
                gaps.append(cur_set)
                cur_set = set()
            prev_idx = idx
        cur_set.add(prev_idx)
        gaps.append(cur_set)
        return gaps


    def find_het_by_resseq(self, resseq):
        """Return a list of heterogens for a residue seq id."""
        modified_het_ids = []
        for res in self:
            if resseq == res.id[1]:
                modified_het_ids.append(res.id)
        return modified_het_ids

    def truncate_missing_terminal(self):
        """Remove the missing residues in reported_res list and the sequence info"""
        bg, end = 1, len(self.can_seq)
        keep_bg = None
        keep_end = None
        terminal_gaps = False
        for gap in self.gaps:
            if bg in gap:
                terminal_gaps = True
                keep_bg = len(gap)
            elif end in gap:
                terminal_gaps = True
                keep_end = -len(gap)
        if not terminal_gaps:
            return
        self.reported_res = sorted(self.reported_res)[keep_bg:keep_end]
        self.can_seq = self.can_seq[keep_bg:keep_end]
        known_seq = self.known_seq
        self.known_seq = known_seq.rstrip(known_seq[keep_end:]).lstrip(known_seq[:keep_bg])

    def is_continuous(self):
        """
        Check if the chain is continuous (no gaps). Missing segments on 
        the C and N terminals will be ignored.
        """
        segments = self.get_segments()
        if len(segments) == 1:
            # If the strucuture's C-N distances are all within the criterion,
            # it is continuous
            return True

        # If not, we also check the sequence here, because after copying gap residues, C-N 
        # distance might be very large before structure minimizations.
        sequence_segments = [
            seq for seq in self.masked_seq.split('-') if len(seq) > 0
        ]
        return len(sequence_segments) == 1

class Heterogens(BaseChain):
    """A chain of heterogens."""
    chain_type = 'Heterogens'
    _res_repr_str = "Molecules"
    def __init__(self, chain_id):
        super().__init__(chain_id)
        self.resnames = None

    def update(self):
        """Update the pdbx_description if only one heterogen exists.
        The description will be assigned directly to that molecule"""
        if len(self) != 1 or self.pdbx_description is None:
            return
        self.child_list[0].pdbx_description = self.pdbx_description

class Ligand(Heterogens):
    """A chain of ligands."""
    chain_type = 'Ligand'
              
class Macrolide(BaseChain):
    chain_type = 'Macrolide'
    
class Oligosaccharide(BaseChain):
    chain_type = 'Oligosaccharide'

class Solvent(BaseChain):
    chain_type = 'Solvent'
    _res_repr_str = "Molecules"
    # The source of the solvent, e.g. PDB (crystallographic) or generated (modeled).
    source: str = None

class CoSolvent(Heterogens):
    chain_type = 'CoSolvent'
    _res_repr_str = "Molecules"


class Ion(BaseChain):
    chain_type = 'Ion'
    _res_repr_str = "Ions"
    source: str = None

class Glycosylation(BaseChain):
    chain_type = 'Glycosylation'

class NucleosidePhosphate(Heterogens):
    chain_type = 'NucleosidePhosphate'


class MaskedSeq(Seq):
    """
    A sequence masked with '-' for any missing residues for visualization purposes.
    The masked sequence is constructed from a PolymerChain class where missing 
    residues are reported. The sequence that is missing will be printed in red 
    if the show() method is called.
    """
    RED = '\033[91m'
    ENDC = '\033[0m'

    def __init__(self, missing_res_ids, can_seq, seq_start, length=None):
        self.color_coded_seq = ''
        masked_seq = ''
        for i, code in enumerate(can_seq, start=seq_start):
            if i in missing_res_ids:
                self.color_coded_seq += f"{self.RED}{code}{self.ENDC}"
                masked_seq += '-'
            else:
                self.color_coded_seq += code
                masked_seq += code
        super().__init__(masked_seq, length)
        self.is_matched = self.check_with_can_seq(can_seq)

    def check_with_can_seq(self, can_seq):
        """Check if the MaskedSequence matches with the canonical sequence for the 
        present residues. Unidentified residues on canonical sequence will be 
        skipped for the check."""
        msg = 'Canonical sequence and masked sequence mismatch!'
        if len(self) != len(can_seq):
            warnings.warn(msg)
            return False
        for masked_char, char in zip(self, can_seq):
            if masked_char != char and masked_char != '-' and char != 'X':
                warnings.warn(msg)
                return False
        return True

    def show(self):
        """Display the entire sequence where the missing residues are colored
        in red."""
        print(self.color_coded_seq)


def convert_chain(chain: _Chain):
    """Convert a Biopython Chain class to general Chain"""
    pchain = Chain(chain.id)
    pchain.set_parent(chain.parent)
    for res in chain:
        res = res.copy()
        pchain.add(res)
    chain.detach_parent()
    return pchain


from Bio.Seq import Seq
from Bio.Data.PDBData import protein_letters_1to3, protein_letters_3to1
from crimm.Modeller import ParameterLoader, ResidueTopologySet
from crimm.StructEntities.Chain import PolymerChain
from crimm.StructEntities.TopoDefinitions import ResidueDefinition

class SeqChainGenerator:
    """Generate a polymer chain from a sequence string"""
    def __init__(self) -> None:
        self.sequence = None
        self.chain_type = None
        self.chain_id = None
        self.topo_definitions = None
        self.params = None
        self._1to3 = None
        self._3to1 = None
        self._report_res = None
        self.resnames = None
        self.built_residues = None

    def _set_topo_definitions(self, definition_type: str):
        self.topo_definitions = ResidueTopologySet(definition_type)
        self.params = ParameterLoader(definition_type)
        self.params.fill_ic(self.topo_definitions, preserve=True)

    def set_chain_type(self, chain_type: str):
        chain_type = chain_type.lower()
        if chain_type in ('protein', 'peptide', 'polypeptide', 'polypeptide(l)'):
            chain_type = 'Polypeptide(L)'
            self._1to3 = protein_letters_1to3
            self._3to1 = protein_letters_3to1
            self._set_topo_definitions('protein')
        elif chain_type in ('rna', 'polyribonucleotide'):
            chain_type = 'Polyribonucleotide'
            self._3to1 = ResidueDefinition.na_3to1
            self._1to3 = ResidueDefinition.na_1to3
            self._set_topo_definitions('nucleic')
        elif chain_type in ('dna', 'polydeoxyribonucleotide'):
            chain_type = 'Polydeoxyribonucleotide'
            raise NotImplementedError('DNA is not supported yet.')
            self._3to1 = ResidueDefinition.na_3to1
            self._1to3 = ResidueDefinition.na_1to3
            self._set_topo_definitions('nucleic')
        else:
            raise ValueError(
                f'Invalid chain type \'{chain_type}\'! '
                f'Valid chain types are: Protein, RNA'
            )
        self.chain_type = chain_type

    def set_sequence(self, sequence: str, chain_type: str):
        self.set_chain_type(chain_type)
        self.resnames = []
        for one_letter in sequence:
            if one_letter not in self._1to3:
                raise ValueError(
                    f'Invalid one-letter code \'{one_letter}\' found '
                    f'in sequence: {sequence}'
                )
            self.resnames.append(self._1to3[one_letter])
        self.sequence = Seq(sequence)
        self._create_reported_residues()

    def set_three_letter_sequence(self, sequence3: str, chain_type: str):
        self.set_chain_type(chain_type)
        self.resnames = sequence3.split()
        sequence = ''
        for resname in self.resnames:
            if resname not in self._3to1:
                raise ValueError(
                    f'Invalid three-letter code \'{resname}\' found '
                    f'in sequence: {sequence3}'
                )
            sequence += self._3to1[resname]
        self.sequence = Seq(sequence)
        self._create_reported_residues()

    def _create_reported_residues(self):
        if self.chain_type == 'Polyribonucleotide' or 'Polydeoxyribonucleotide':
            # RNA and DNA are reported in one-letter code for PDBx format
            PDB_resnames = [self._3to1[resname] for resname in self.resnames]
        else:
            # Protein is reported in three-letter code for PDBx format
            PDB_resnames = self.resnames
        self._report_res = [
            (i+1, resname) for i, resname in enumerate(PDB_resnames)
        ]
        
    def _create_initial_residue(self):
        resname = self.resnames[0]
        res_def = self.topo_definitions[resname]
        res = res_def.create_residue(resseq = 1)
        return res
    
    def create_chain(self, chain_id: str = 'A'):
        self.chain_id = chain_id
        init_residue = self._create_initial_residue()
        self.built_residues = {1: init_residue}
        cur_coord_dict = self.topo_definitions[self.resnames[0]].standard_coord_dict
        for resseq, resname in enumerate(self.resnames[1:], start=2):
            last_res = self.built_residues[resseq-1]
            last_def = self.topo_definitions[last_res.resname]
            cur_def = self.topo_definitions[resname]
            if cur_def.standard_coord_dict is None:
                cur_def.standard_coord_dict = cur_def.build_standard_coord_from_ic_table()
            cur_coord_dict = self._create_neighbor_coord_dict(last_def, cur_def, cur_coord_dict)
            cur_res = cur_def.create_residue_from_coord_dict(
                cur_coord_dict, resseq = resseq
            )
            self.built_residues[resseq] = (cur_res)

        chain = PolymerChain(
            chain_id=self.chain_id,
            chain_type=self.chain_type,
            author_chain_id=self.chain_id,
            entity_id=1,
            canon_sequence=self.sequence,
            known_sequence=self.sequence,
            reported_res = self._report_res
        )

        for res in self.built_residues.values():
            chain.add(res)

        return chain

    def _find_nei_ic_entry(self, topo_definition):
        for atoms in topo_definition.ic:
            nei_atoms_next = []
            cur_atoms = []
            for i, atom in enumerate(atoms):
                if atom.startswith('+'):
                    nei_atoms_next.append((i, atom))
                else:
                    cur_atoms.append((i, atom))
            if len(nei_atoms_next) == 2:
                return cur_atoms, nei_atoms_next

    def _find_init_coord_for_nei(self, standard_coord_dict, atoms, nei_atoms):
        (i, a1), (j, a2) = atoms
        (k, a3), (l, a4) = nei_atoms
        if i < j < k < l:
            atom_names = (a2, a3, a4)
        elif l < k < j < i:
            atom_names = (a4, a3, a2)
        coords = (standard_coord_dict[atom_name] for atom_name in atom_names)
        nei_atom_names = []
        for atom_name in atom_names:
            if atom_name.startswith('+'):
                nei_atom_names.append(atom_name[1:])
            else:
                nei_atom_names.append('-'+atom_name)
        return dict(zip(nei_atom_names, coords))

    def _create_neighbor_coord_dict(self, last_def, cur_def, cur_coord_dict):
        atoms, nei_atoms = self._find_nei_ic_entry(last_def)
        init_coord_dict = self._find_init_coord_for_nei(cur_coord_dict, atoms, nei_atoms)
        new_coord_dict = cur_def.build_standard_coord_from_ic_table(
            init_coord_dict=init_coord_dict
        )
        return new_coord_dict



from string import ascii_uppercase
import warnings
import pathlib
from Bio.PDB.PDBParser import PDBParser as _PDBParser
from Bio.Data.PDBData import (
    protein_letters_3to1, protein_letters_3to1_extended,
    nucleic_letters_3to1, nucleic_letters_3to1_extended
)
from crimm.IO.StructureBuilder import StructureBuilder
from crimm.StructEntities.Chain import Solvent, Heterogens, PolymerChain, Chain, Ion
from crimm.Data.components_dict import CHARMM_PDB_ION_NAMES

protein_letters_3to1.update({'HSD': 'H', 'HSE': 'H', 'HSP': 'H'})
def check_chain_type(residues, resname_lookup, extended_lookup=None):
    for res in residues:
        if res.resname not in resname_lookup:
            # Check if the residue is a heterogen and if it is in the extended
            # lookup table, heterogen flag will be changed to 'H_{resname}'
            if extended_lookup and (res.resname in extended_lookup):
                hetflag, resseq, icode = res.get_id()
                hetflag = f'H_{res.resname}'
                res.id = (hetflag, resseq, icode)
                warnings.warn(f"Heterogen flag for {res.get_id()} corrected.")
                continue
            return False
    return True

def separate_solvent(residues):
    het = []
    water = []
    for res in residues:
        if res.resname == 'HOH' or res.resname == 'TIP3':
            water.append(res)
        else:
            het.append(res)
    return het, water

def find_chain_type(chain):
    residues = []
    het_residues = []
    for res in chain:
        hetflag, resseq, icode = res.get_id()
        if hetflag == " ":
            residues.append(res)
        else:
            het_residues.append(res)
    if residues:
        if check_chain_type(
            residues, protein_letters_3to1, protein_letters_3to1_extended
        ):
            return [('Polypeptide(L)', chain.id, residues)]
        elif check_chain_type(
            residues, nucleic_letters_3to1, nucleic_letters_3to1_extended
        ):
            return [('Polyribonucleotide', chain.id, residues)]
        elif check_chain_type(
            residues, CHARMM_PDB_ION_NAMES, None
        ):
            return [('Ion', chain.id, residues)]
        else:
            warnings.warn(
                f'Chain type cannot be determined for {chain.get_full_id()}'
            )
            return [('Chain', chain.id, residues)]
    else:
        het, water = separate_solvent(het_residues)
        return [('Heterogens', chain.id, het), ('Solvent', chain.id, water)]

def convert_chains(chains):
    organized_residues = []
    for chain in chains:
        organized_residues.extend(find_chain_type(chain))

    chain_sort_dict = {
        'Polypeptide(L)': 0,
        'Polyribonucleotide': 1,
        'Chain': 2, # 'Chain' is a catch-all for 'unknown
        'Heterogens': 3,
        'Ion': 4,
        'Solvent': 5
    }

    new_chains = []
    for i, (chain_type, auth_chain_id, residues) in enumerate(
        sorted(organized_residues, key=lambda x: chain_sort_dict[x[0]])
    ):
        if len(residues) == 0:
            continue
        new_chain_id = ascii_uppercase[i]
        if chain_type in ('Polypeptide(L)', 'Polyribonucleotide'):
            new_chain = PolymerChain(
                chain_id = new_chain_id,
                entity_id = None,
                author_chain_id = auth_chain_id,
                chain_type = chain_type,
                known_sequence = '',
                canon_sequence = '',
                reported_res = [],
                reported_missing_res = []
            )
        elif chain_type == 'Chain':
            new_chain = Chain(
                chain_id = new_chain_id,
            )
        elif chain_type == 'Heterogens':
            new_chain = Heterogens(new_chain_id)
        elif chain_type == 'Ion':
            new_chain = Ion(new_chain_id)
        else:
            new_chain = Solvent(new_chain_id)
        for res in residues:
            res.detach_parent()
            if res.resname == 'ILE' and 'CD' in res:
                # ILE could have a CD atom (CHARMM convention) 
                # that is not in the standard PDB format
                atom = res['CD']
                res.detach_child('CD')
                atom.name = 'CD1'
                atom.id = 'CD1'
                atom.fullname = ' CD1'
                res.add(atom)
            new_chain.add(res)
        new_chains.append(new_chain)
    return new_chains

class PDBParser(_PDBParser):
    """PDBParser that returns a Structure with determined chain types."""
    def __init__(
            self, 
            first_model_only = True,
            include_solvent = True,
            strict_parser=True, 
            get_header=False, 
            QUIET=False
        ):
        structure_builder = StructureBuilder()
        self.first_model_only = first_model_only
        self.include_solvent = include_solvent
        PERMISSIVE = not strict_parser
        super().__init__(
            PERMISSIVE, get_header, structure_builder, QUIET, is_pqr=False
        )

    def _get_structure(self, filepath, structure_id=None):
        """Return the structure contained in file."""
        if structure_id is None:
            structure_id = pathlib.Path(filepath).stem
        structure = super().get_structure(structure_id, filepath)
        if self.first_model_only:
            structure.child_list = structure.child_list[:1]
            structure.child_dict = {c.id: c for c in structure.child_list}
        for model in structure:
            new_chains = convert_chains(model.child_list)
            if not self.include_solvent:
                new_chains = [c for c in new_chains if c.chain_type != 'Solvent']
            for chain in new_chains:
                chain.set_parent(model)
            model.child_list = new_chains
            model.child_dict = {c.id: c for c in new_chains}
        return structure

    def get_structure(self, filepath, structure_id=None):
        """Return the structure contained in file."""
        if self.QUIET:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return self._get_structure(filepath, structure_id)
        return self._get_structure(filepath, structure_id)
    


"""This module contains the Solvator class, which solvates a Structure, Model,
or Chain level entity with water molecules.
"""
import os
import warnings
import numpy as np
import math
from typing import Optional, Tuple
from dataclasses import dataclass
import random
from random import choices
from scipy.spatial import KDTree
from crimm import Data
from crimm.Modeller.CoordManipulator import CoordManipulator
from crimm.StructEntities.Atom import Atom
from crimm.StructEntities.Residue import Residue
from crimm.StructEntities.Model import Model
from crimm.StructEntities.Chain import Solvent, Ion
from crimm.Data.components_dict import CHARMM_PDB_ION_NAMES
from crimm.Modeller.TopoLoader import TopologyGenerator


WATER_COORD_PATH = os.path.join(os.path.dirname(Data.__file__), 'water_coords.npy')
BOXWIDTH=18.662 # water unit cube width

# Physical constants for ion calculations
WATER_MOLARITY = 55.5  # Molar concentration of pure water at 298K (often approximated as 56)
AVOGADRO = 6.02214076e23  # mol^-1
NM3_TO_L = 1e-24  # nm³ to liters conversion
A3_TO_NM3 = 1e-3  # Å³ to nm³ conversion
WATERS_PER_NM3 = 33.3  # Approximate number of water molecules per nm³ at 298K

# Crystal type definition
@dataclass
class CrystalType:
    """CHARMM crystal type definition.

    Attributes
    ----------
    name : str
        Display name for the crystal type
    charmm_name : str
        CHARMM 4-character code (CUBI, OCTA, RHDO, etc.)
    dof : int
        Degrees of freedom for lattice optimization
    volume_factor : float
        Volume relative to cubic box with same max dimension (for isotropic types)
    angles : Tuple[float, float, float]
        Crystal angles (alpha, beta, gamma) in degrees
    supports_solvation : bool
        Whether water box generation is supported for this type
    """
    name: str
    charmm_name: str
    dof: int
    volume_factor: float
    angles: Tuple[float, float, float]
    supports_solvation: bool

# All CHARMM crystal types (from ~/software/charmm/doc/crystl.info)
CRYSTAL_TYPES = {
    'cube': CrystalType(
        name='Cubic',
        charmm_name='CUBI',
        dof=1,
        volume_factor=1.0,
        angles=(90.0, 90.0, 90.0),
        supports_solvation=True
    ),
    'octa': CrystalType(
        name='Truncated Octahedron',
        charmm_name='OCTA',
        dof=1,
        volume_factor=0.77,  # (4*sqrt(3)/9) ≈ 0.77
        angles=(109.4712206344907, 109.4712206344907, 109.4712206344907),
        supports_solvation=True
    ),
    'rhdo': CrystalType(
        name='Rhombic Dodecahedron',
        charmm_name='RHDO',
        dof=1,
        volume_factor=0.707,  # sqrt(0.5) ≈ 0.707
        angles=(60.0, 90.0, 60.0),
        supports_solvation=True
    ),
    'ortho': CrystalType(
        name='Orthorhombic',
        charmm_name='ORTH',
        dof=3,
        volume_factor=1.0,  # Variable, depends on a,b,c
        angles=(90.0, 90.0, 90.0),
        supports_solvation=True
    ),
    'tetra': CrystalType(
        name='Tetragonal',
        charmm_name='TETR',
        dof=2,
        volume_factor=1.0,  # Variable, depends on a,c
        angles=(90.0, 90.0, 90.0),
        supports_solvation=True
    ),
    'hexa': CrystalType(
        name='Hexagonal',
        charmm_name='HEXA',
        dof=2,
        volume_factor=0.866,  # sqrt(0.75) ≈ 0.866
        angles=(90.0, 90.0, 120.0),
        supports_solvation=True
    ),
    'mono': CrystalType(
        name='Monoclinic',
        charmm_name='MONO',
        dof=4,
        volume_factor=1.0,
        angles=(90.0, 90.0, 90.0),  # beta varies
        supports_solvation=False
    ),
    'tric': CrystalType(
        name='Triclinic',
        charmm_name='TRIC',
        dof=6,
        volume_factor=1.0,
        angles=(90.0, 90.0, 90.0),  # all vary
        supports_solvation=False
    ),
    'rhomb': CrystalType(
        name='Rhombohedral',
        charmm_name='RHOM',
        dof=2,
        volume_factor=1.0,  # Variable
        angles=(60.0, 60.0, 60.0),  # all equal, <120
        supports_solvation=False
    ),
}

# Volume factors for different box geometries (relative to cubic)
# Kept for backward compatibility
BOX_VOLUME_FACTORS = {k: v.volume_factor for k, v in CRYSTAL_TYPES.items()}

# Ion valence table - only ions defined in crimm/Data/toppar/water_ions.str
ION_VALENCES = {
    # Alkali metals (+1)
    'LIT': +1,   # Lithium
    'SOD': +1,   # Sodium
    'POT': +1,   # Potassium
    'RUB': +1,   # Rubidium
    'CES': +1,   # Cesium
    # Alkaline earth (+2)
    'MG': +2,    # Magnesium
    'CAL': +2,   # Calcium
    'BAR': +2,   # Barium
    # Transition metals (defined in toppar)
    'ZN2': +2,   # Zinc
    'CD2': +2,   # Cadmium
    # Anions
    'CLA': -1,   # Chloride
    'OH': -1,    # Hydroxide
}

# Topology definitions for water and ions
_topo = TopologyGenerator()
_topo_def, _params = _topo.load_residue_definitions('Solvent')

TIP3_def = _topo_def['TIP3']
# Create new Atom instances without coordinates
OH2 = TIP3_def['OH2'].create_new_atom()
H1 = TIP3_def['H1'].create_new_atom()
H2 = TIP3_def['H2'].create_new_atom()

## TODO: deal with atom serial number > 99999 for larger structures (e.g. 1A8I)
class Solvator:
    """Solvates a Structure, Model, or Chain level entity with water molecules.
    The solvated entity will be returned as a Model level entity. The solvated
    entity will be centered in a cubic box with side length equal to the
    maximum dimension of the entity plus the cutoff distance. (i.e., Coordinates 
    will be oriented using CoordManipulator.orient_coords() before solvation.)
    The solvcut distance is the distance from the solute at which water
    molecules will be removed. The solvcut distance is used to remove water 
    molecules that are too close to the solute. 
    If altloc atoms exist in the entity, the first altloc atoms will be used to
    determine water molecules location during solvation.

    Parameters
    ----------
    entity : Structure, Model, or Chain level entity
        The entity to solvate. If a Structure level entity is provided, the
        first Model will be solvated. If a Model level entity is provided, all 
        chains in the model will be solvated. If a Chain level entity is 
        provided, the chain will be solvated. The entity is modified in place.

    Examples
    --------
    >>> from crimm import fetch_rcsb
    >>> from crimm.Modeller.Solvator import Solvator

    >>> model = fetch_rcsb('5igv')
    >>> solvator = Solvator(model)
    >>> water_chains = solvator.solvate(
            cutoff=8.0, solvcut=2.1, remove_existing_water=True
        )
    >>> solvent_chains = [
        chain for chain in model if chain.chain_type == 'Solvent'
    ]

    Note that water chains are named W[A-Z]. All waters are placed in a single chain
    (PSF/CRD extended format supports residue numbers up to 10 digits).
    >>> water_chains
    [<Solvent id=WA Residues=12485>]
    >>> solvator.water_box_coords.shape # shape in (N waters, 3 atoms, 3 coords)
    (12485, 3, 3)
    >>> chain = structure[1]['A'] # get chain A from the first model
    >>> solvated_chain = solvator.solvate(chain)

    More water molecules are added to solvate the chain since the ligands are
    not included in the solvation process
    >>> solvator.water_box_coords.shape 
    (12531, 3, 3)

    """

    alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    available_ions = CHARMM_PDB_ION_NAMES
    def __init__(self, entity) -> None:
        if entity.level == 'S':
            self.model = entity.models[0]
        elif entity.level == 'M':
            self.model = entity
        elif entity.level == 'C':
            self.model = Model(1)
            self.model.add(entity)
        else:
            raise ValueError(
                'Solvator can only take Structure, Model, or Chain level entities'
            )
        
        self.cutoff = None
        self.solvcut = None
        self.coords = None
        self.box_dim = None
        self.box_dims = None  # For orthorhombic: (a, b, c)
        self.water_box_coords = None
        # unit of pre-equilibrated cube of water molecules (18.662 A each side)
        self.water_unit_coords = np.load(WATER_COORD_PATH)
        self._topo_loader = getattr(self.model, 'topology_loader', None)
        self.box_type = None
        self.orient_method = None
        self.crystal_type = None  # CrystalType object
        
    def get_model(self):
        return self.model

    def suggest_optimal_crystal(
            self,
            cutoff: float = 9.0,
            candidates: list = None
        ) -> dict:
        """Suggest optimal crystal type based on molecule geometry.

        Analyzes the molecule's bounding box dimensions after PCA orientation
        and calculates the solvent volume for each candidate crystal type.
        Returns the type that minimizes total solvent volume.

        The function simulates the orientation that would be applied during
        solvation to provide accurate volume estimates.

        Parameters
        ----------
        cutoff : float, optional
            Minimum distance from solute to box edge in Angstroms (default: 9.0)
        candidates : list, optional
            List of crystal types to consider.
            Default: ['cube', 'octa', 'rhdo', 'ortho']

        Returns
        -------
        dict
            Dictionary with:
            - 'recommended': str, the optimal crystal type
            - 'reason': str, explanation for the recommendation
            - 'volumes': dict, volume comparison for all candidates (in Å³)
            - 'molecule_shape': str, 'globular', 'elongated', or 'intermediate'
            - 'extents': tuple, molecule dimensions after orientation (sorted: largest first)
            - 'aspect_ratios': dict, shape descriptors

        Examples
        --------
        >>> solvator = Solvator(model)
        >>> result = solvator.suggest_optimal_crystal(cutoff=10.0)
        >>> print(f"Recommended: {result['recommended']}")
        >>> print(f"Reason: {result['reason']}")
        >>> solvator.solvate(cutoff=10.0, box_type=result['recommended'])
        """
        if candidates is None:
            candidates = ['cube', 'octa', 'rhdo', 'ortho']

        # Validate candidates
        for c in candidates:
            if c not in CRYSTAL_TYPES:
                raise ValueError(f"Unknown crystal type: {c}")
            if not CRYSTAL_TYPES[c].supports_solvation:
                raise ValueError(f"Crystal type '{c}' does not support solvation")

        # Extract coordinates from model EXCLUDING water and ion chains
        # This matches solvation behavior where waters are removed before orientation
        coords = self._extract_coords(self.model, non_solvent_only=True)

        # Simulate orient_coords_ortho: PCA with axis sorting by singular values
        # This matches exactly what happens during solvation with ortho orientation
        centered = coords - coords.mean(axis=0)
        U, S, Vt = np.linalg.svd(centered, full_matrices=False)
        
        # Sort principal components by decreasing singular values (largest variance first)
        sort_idx = np.argsort(S)[::-1]
        rotation = Vt[sort_idx].T
        
        # Ensure right-handed coordinate system
        if np.linalg.det(rotation) < 0:
            rotation[:, -1] *= -1
        
        # Apply rotation to get oriented coordinates
        oriented_coords = centered @ rotation
        
        # Get extents after orientation (now sorted: largest on X, smallest on Z)
        extents = np.ptp(oriented_coords, axis=0)
        
        # These are already sorted by the SVD-based orientation
        # extents[0] >= extents[1] >= extents[2] (approximately)
        sorted_extents = extents  # Already sorted by orientation
        
        # Calculate aspect ratios
        aspect_xy = sorted_extents[0] / sorted_extents[1] if sorted_extents[1] > 0 else 1.0
        aspect_xz = sorted_extents[0] / sorted_extents[2] if sorted_extents[2] > 0 else 1.0

        # Classify molecule shape
        if aspect_xy < 1.3 and aspect_xz < 1.5:
            shape = 'globular'
        elif aspect_xz > 2.5:
            shape = 'elongated'
        else:
            shape = 'intermediate'

        # Calculate volumes for each candidate
        volumes = {}
        for crystal_type in candidates:
            if crystal_type == 'ortho':
                # Orthorhombic: use actual extents from orientation
                dims = sorted_extents + 2 * cutoff
                vol = dims[0] * dims[1] * dims[2]
            elif crystal_type == 'tetra':
                # Tetragonal: a=b (max of X,Y), c (Z)
                a = max(sorted_extents[0], sorted_extents[1]) + 2 * cutoff
                c = sorted_extents[2] + 2 * cutoff
                vol = a * a * c
            elif crystal_type == 'hexa':
                # Hexagonal: a=b (max of X,Y), c (Z with smallest extent)
                # Volume = sqrt(3)/2 * a² * c ≈ 0.866 * a² * c
                a = max(sorted_extents[0], sorted_extents[1]) + 2 * cutoff
                c = sorted_extents[2] + 2 * cutoff
                vol = 0.866 * a * a * c
            else:
                # Isotropic types (cube, octa, rhdo) use max extent
                box_dim = sorted_extents.max() + 2 * cutoff
                vol_factor = CRYSTAL_TYPES[crystal_type].volume_factor
                vol = vol_factor * (box_dim ** 3)
            volumes[crystal_type] = vol

        # Find minimum volume
        recommended = min(volumes, key=volumes.get)
        min_vol = volumes[recommended]
        cube_vol = volumes.get('cube', min_vol)

        # Generate reason with savings percentage
        savings = (cube_vol - min_vol) / cube_vol * 100 if cube_vol > 0 else 0

        if recommended == 'ortho':
            if shape == 'elongated':
                reason = (f"Molecule is elongated (aspect ratio {aspect_xz:.1f}:1). "
                         f"Orthorhombic box saves {savings:.0f}% volume vs cubic.")
            else:
                reason = (f"Orthorhombic fits molecule dimensions best, "
                         f"saving {savings:.0f}% volume vs cubic.")
        elif recommended in ('octa', 'rhdo'):
            reason = (f"Molecule is {shape}. {CRYSTAL_TYPES[recommended].name} "
                     f"saves {savings:.0f}% volume vs cubic.")
        else:
            reason = f"{CRYSTAL_TYPES[recommended].name} minimizes solvent volume ({savings:.0f}% savings)."

        return {
            'recommended': recommended,
            'reason': reason,
            'volumes': volumes,
            'molecule_shape': shape,
            'extents': tuple(sorted_extents),
            'aspect_ratios': {'xy': aspect_xy, 'xz': aspect_xz}
        }

    def solvate(
            self, cutoff=9.0, solvcut = 2.10,
            remove_existing_water = True,
            remove_existing_ions = False,
            orient_coords = True,
            box_type = 'cube',
            orient_method = None,
            box_dims = None

        ) -> list:
        """Solvate the entity with a water box.

        The solvated entity will be centered in a water box with dimensions
        based on the chosen crystal type. Coordinates will be oriented using
        CoordManipulator before solvation. The solvcut distance is used to
        remove water molecules that are too close to the solute.

        The model is modified in place - water chains are added directly to the
        model. The returned list contains the added water chains.

        Parameters
        ----------
        cutoff : float, optional
            The distance from the solute to the edge of the box (default: 9.0 Å)
        solvcut : float, optional
            The distance from the solute at which water molecules will be
            removed (default: 2.10 Å)
        remove_existing_water : bool, optional
            If True, remove existing water molecules from the entity (default: True)
        remove_existing_ions : bool, optional
            If True, remove existing ions from the entity (default: False)
        orient_coords : bool, optional
            If True, orient coordinates before solvation (default: True)
        box_type : str, optional
            The shape of the water box. Supported types:
            - 'cube': Cubic box (default)
            - 'octa': Truncated octahedron (~23% volume savings)
            - 'rhdo': Rhombic dodecahedron (~29% volume savings)
            - 'ortho': Orthorhombic (rectangular, optimal for elongated molecules)
            - 'tetra': Tetragonal (a=b, c variable)
            - 'hexa': Hexagonal prism
            Use suggest_optimal_crystal() to get recommendations.
        orient_method : str, optional
            The method to orient the coordinates before solvation:
            - None or 'default': Standard orientation (major axis along X)
            - 'octa': PCA-based orientation for octahedral/rhdo boxes
        box_dims : tuple, optional
            For orthorhombic/tetragonal boxes, specify (a, b, c) dimensions.
            If None, dimensions are calculated from molecule extents + cutoff.

        Returns
        -------
        list
            List of Solvent chains added to the model.

        See Also
        --------
        suggest_optimal_crystal : Recommend optimal crystal type based on geometry
        CRYSTAL_TYPES : Dictionary of all supported crystal types
        """
        # Validate crystal type
        if box_type not in CRYSTAL_TYPES:
            raise ValueError(
                f"Unknown box_type '{box_type}'. "
                f"Supported types: {list(CRYSTAL_TYPES.keys())}"
            )

        self.crystal_type = CRYSTAL_TYPES[box_type]

        if not self.crystal_type.supports_solvation:
            raise ValueError(
                f"Crystal type '{box_type}' does not support water box generation. "
                f"Solvation-compatible types: "
                f"{[k for k, v in CRYSTAL_TYPES.items() if v.supports_solvation]}"
            )

        self.cutoff = cutoff
        self.solvcut = solvcut
        self.box_type = box_type
        self.orient_method = orient_method

        # Set default orientation method based on box type
        # Each crystal type benefits from a specific orientation strategy
        if self.orient_method is None:
            if self.box_type in ("octa", "rhdo", "cube"):
                # PCA minimizes bounding box for isotropic geometries
                self.orient_method = "octa"
            elif self.box_type in ("ortho", "tetra"):
                # Sort axes by extent: largest→X, medium→Y, smallest→Z
                self.orient_method = "ortho"
            elif self.box_type == "hexa":
                # Hexagonal: put flattest dimension on Z (prism height)
                self.orient_method = "hexa"
            else:
                self.orient_method = "default"
        
        # Initialize solvation info storage
        if not hasattr(self.model, '_solvation_info'):
            self.model._solvation_info = {}

        if remove_existing_water:
            self.remove_existing_water(self.model)
            self.model._solvation_info['preserved_waters'] = 0
        else:
            # Convert existing waters to TIP3 format for CHARMM compatibility
            n_preserved_waters = self._convert_existing_waters_to_tip3(self.model)
            self.model._solvation_info['preserved_waters'] = n_preserved_waters

        if remove_existing_ions:
            self.remove_existing_ions(self.model)
            self.model._solvation_info['preserved_ions'] = 0
        else:
            # preserve existing ions (convert names if needed)
            n_preserved_ions = self._convert_existing_ions(self.model)
            self.model._solvation_info['preserved_ions'] = n_preserved_ions

        if len(self.model.chains) == 0:
            raise ValueError('No chains in model to solvate')
        if orient_coords:
            coorman = CoordManipulator()
            coorman.load_entity(self.model)
            apply_parent = self.model.parent is not None

            if self.orient_method == "octa":
                coorman.orient_coords_octa(apply_to_parent=apply_parent)
                warnings.warn("Using PCA orientation for solvation.", UserWarning)
            elif self.orient_method == "ortho":
                coorman.orient_coords_ortho(apply_to_parent=apply_parent)
                warnings.warn("Using orthorhombic orientation for solvation.", UserWarning)
            elif self.orient_method == "hexa":
                coorman.orient_coords_hexa(apply_to_parent=apply_parent)
                warnings.warn("Using hexagonal orientation for solvation.", UserWarning)
            else:
                coorman.orient_coords(apply_to_parent=apply_parent)
                warnings.warn("Using default orientation for solvation.", UserWarning)

            warnings.warn(
                'Orienting coordinates before solvation. This may change the '
                'atom coordinates of the entity in the structure.',
                UserWarning
            )
        self.coords = self._extract_coords(self.model, non_solvent_only=True)
        # Box dimension cannot be smaller than the existing solvent molecule extents
        cur_solvent_dim = self._extract_coords(self.model, non_solvent_only=False)
        min_box_dim = np.ptp(cur_solvent_dim, axis=0)

        # Calculate box dimensions based on crystal type
        extents = np.ptp(self.coords, axis=0)
        
        if self.box_type in ('cube', 'octa', 'rhdo'):
            # Isotropic boxes: use max extent
            self.box_dim = extents.max() + 2 * self.cutoff
            self.box_dims = (self.box_dim, self.box_dim, self.box_dim)
        elif self.box_type == 'ortho':
            # Orthorhombic: use actual extents for each dimension
            if box_dims is not None:
                self.box_dims = tuple(box_dims)
            else:
                self.box_dims = tuple(extents + 2 * self.cutoff)
            self.box_dim = max(self.box_dims)  # For compatibility
        elif self.box_type == 'tetra':
            # Tetragonal: a=b, c can be different
            if box_dims is not None:
                self.box_dims = tuple(box_dims)
            else:
                a = max(extents[0], extents[1]) + 2 * self.cutoff
                c = extents[2] + 2 * self.cutoff
                self.box_dims = (a, a, c)
            self.box_dim = max(self.box_dims)
        elif self.box_type == 'hexa':
            # Hexagonal: a=b, c can be different
            if box_dims is not None:
                self.box_dims = tuple(box_dims)
            else:
                a = max(extents[0], extents[1]) + 2 * self.cutoff
                c = extents[2] + 2 * self.cutoff
                self.box_dims = (a, a, c)
            self.box_dim = max(self.box_dims)
        else:
            # Fallback for any other type
            self.box_dim = extents.max() + 2 * self.cutoff
            self.box_dims = (self.box_dim, self.box_dim, self.box_dim)
        if self.box_dims[0] < min_box_dim[0] or \
           self.box_dims[1] < min_box_dim[1] or \
           self.box_dims[2] < min_box_dim[2]:
            raise ValueError(
                f'Calculated box dimensions {self.box_dims}Å are smaller than existing '
                f'solvent molecule extents {min_box_dim}Å. Increase cutoff or provide '
                'larger box_dims for orthorhombic/tetragonal boxes.'
            )
        return self._solvate_model()


    def _extract_coords(self, entity, non_solvent_only: bool = False) -> np.ndarray:
        """Extracts coordinates from entity. If any altloc atoms are present,
        only the first altloc atoms will be included in the returned array.
        Parameters
        ----------
        entity : Structure, Model, or Chain level entity
            The entity from which to extract coordinates.
        non_solvent_only : bool, optional
            If True and entity is Model level, exclude water and ion chains.
            Default is False.
        Returns
        -------
        np.ndarray
            Array of coordinates.
        """
        # get non-solvent atoms only if specified
        if non_solvent_only and entity.level == 'M':
            atoms = []
            for chain in entity:
                chain_type = getattr(chain, 'chain_type', None)
                # Skip water and ion chains
                if chain_type in ('Solvent', 'Ion'):
                    continue
                for atom in chain.get_atoms(include_alt=False):
                    atoms.append(atom)
        else:
            # get all atoms
            atoms = entity.get_atoms(include_alt=False)
        
        coords = [atom.coord for atom in atoms]
        return np.array(coords)

    def remove_existing_water(self, model: Model) -> Model:
        """Removes existing water molecules from the model."""
        remove_list = []
        for chain in model.chains:
            if chain.chain_type == 'Solvent':
                remove_list.append(chain.id)
            elif chain.chain_type == 'Ion' and chain.source == 'generated':
                # Also remove generated ions (not crystallographic)
                remove_list.append(chain.id)
        for chain_id in remove_list:
            warnings.warn(
                f'Removing existing water chain {chain_id} from model',
                UserWarning
            )
            model.detach_child(chain_id)
    
    def remove_existing_ions(self, model: Model) -> Model:
        """Removes existing ions from the model."""
        remove_list = []
        for chain in model.chains:
            if chain.chain_type == 'Ion':
                remove_list.append(chain.id)
        for chain_id in remove_list:
            warnings.warn(
                f'Removing existing ion chain {chain_id} from model',
                UserWarning
            )
            model.detach_child(chain_id)

    def _build_water_hydrogens(self, oxygen_coord: np.ndarray) -> tuple:
        """Build hydrogen positions for a water molecule from oxygen position.

        Uses TIP3P geometry:
        - O-H bond length: 0.9572 Å
        - H-O-H angle: 104.52°

        Parameters
        ----------
        oxygen_coord : np.ndarray
            Oxygen atom coordinates (x, y, z).

        Returns
        -------
        tuple
            (h1_coord, h2_coord) as numpy arrays.
        """
        # TIP3P geometry
        oh_bond = 0.9572  # Å
        hoh_angle = 104.52 * np.pi / 180  # radians

        # Generate random orientation
        # Create a random unit vector for first O-H bond
        random_vec = np.random.randn(3)
        random_vec /= np.linalg.norm(random_vec)

        # First hydrogen
        h1_coord = oxygen_coord + oh_bond * random_vec

        # Second hydrogen: rotate around a perpendicular axis
        # Find a perpendicular vector
        perp = np.cross(random_vec, np.array([1, 0, 0]))
        if np.linalg.norm(perp) < 0.1:
            perp = np.cross(random_vec, np.array([0, 1, 0]))
        perp /= np.linalg.norm(perp)

        # Rotate random_vec by HOH angle around perpendicular axis
        cos_a = np.cos(hoh_angle)
        sin_a = np.sin(hoh_angle)
        # Rodrigues rotation formula
        h2_vec = (random_vec * cos_a +
                  np.cross(perp, random_vec) * sin_a +
                  perp * np.dot(perp, random_vec) * (1 - cos_a))
        h2_coord = oxygen_coord + oh_bond * h2_vec

        return h1_coord, h2_coord
        
    def _convert_existing_waters_to_tip3(self, model: Model) -> int:
        """Convert existing water molecules (HOH/WAT) to CHARMM TIP3 format.

        This:
        - Renames residue: HOH/WAT -> TIP3
        - Renames oxygen: O/OW -> OH2
        - Builds missing hydrogens for crystallographic waters

        Returns
        -------
        int
            Number of crystal waters preserved/converted
        """
        # Standard atom name mappings for water oxygen
        oxygen_names = {'O', 'OW', 'OH2'}

        n_converted = 0
        n_hydrogens_built = 0

        for chain in model.chains:
            if chain.chain_type != 'Solvent':
                continue
            chain.source = 'crystallographic'  # mark as original waters
            for residue in chain.get_residues():
                if residue.resname in ('HOH', 'WAT', 'SOL', 'TIP3'):
                    atoms = list(residue.get_atoms())

                    # Find oxygen atom
                    oxygen_atom = None
                    hydrogen_atoms = []
                    for atom in atoms:
                        if atom.name in oxygen_names or atom.element == 'O':
                            oxygen_atom = atom
                        elif atom.element == 'H':
                            hydrogen_atoms.append(atom)

                    if oxygen_atom is None:
                        # No oxygen found - skip this residue
                        continue

                    # Build missing hydrogens if needed
                    if len(hydrogen_atoms) < 2:
                        h1_coord, h2_coord = self._build_water_hydrogens(oxygen_atom.coord)
                        n_hydrogens_built += (2 - len(hydrogen_atoms))

                        # Get next available serial number
                        max_serial = max(a.serial_number for a in atoms) if atoms else 0

                        if len(hydrogen_atoms) == 0:
                            # Build both hydrogens
                            h1 = H1.copy()
                            h2 = H2.copy()
                            h1.coord = h1_coord
                            h2.coord = h2_coord
                            residue.add(h1)
                            residue.add(h2)
                        elif len(hydrogen_atoms) == 1:
                            h2 = H2.copy()
                            h2.coord = h2_coord
                            residue.add(h2)

                    # Convert residue name to TIP3
                    if residue.resname != 'TIP3':
                        residue.resname = 'TIP3'
                        n_converted += 1

                    # Rename oxygen atom and set topology definition
                    if oxygen_atom.name != 'OH2':
                        oxygen_atom.name = 'OH2'
                        oxygen_atom.fullname = ' OH2'
                    # Ensure oxygen has proper topology definition (charge=-0.834, type=OT)
                    oxygen_atom.topo_definition = OH2.topo_definition

        # Count total preserved waters (converted + already TIP3)
        n_preserved = 0
        for chain in model.chains:
            if chain.chain_type == 'Solvent':
                n_preserved = len(list(chain.get_residues()))
                break

        if n_converted > 0:
            warnings.warn(
                f'Converted {n_converted} existing water molecules to TIP3 format',
                UserWarning
            )
        if n_hydrogens_built > 0:
            warnings.warn(
                f'Built {n_hydrogens_built} missing hydrogens for crystallographic waters',
                UserWarning
            )

        return n_preserved

    def _convert_existing_ions(self, model: Model) -> int:
        """Convert existing ions to CHARMM-compatible names.

        Only converts ions that are defined in crimm/Data/toppar/water_ions.str.
        Maps common PDB ion names to CHARMM residue names.

        Returns
        -------
        int
            Number of crystal ions preserved
        """
        # Map PDB ion names to CHARMM names (only toppar-defined ions)
        ion_name_map = {
            # Alkali metals
            'LI': 'LIT', 'LI+': 'LIT',
            'NA': 'SOD', 'NA+': 'SOD', 'NAI': 'SOD',
            'K': 'POT', 'K+': 'POT',
            'RB': 'RUB', 'RB+': 'RUB',
            'CS': 'CES', 'CS+': 'CES',
            # Alkaline earth
            'MG': 'MG', 'MG2+': 'MG', 'MG2': 'MG',
            'CA': 'CAL', 'CA2+': 'CAL', 'CA2': 'CAL',
            'BA': 'BAR', 'BA2+': 'BAR', 'BA2': 'BAR',
            # Transition metals (defined in toppar)
            'ZN': 'ZN2', 'ZN2+': 'ZN2',
            'CD': 'CD2', 'CD2+': 'CD2',
            # Anions
            'CL': 'CLA', 'CL-': 'CLA',
        }

        n_converted = 0
        n_preserved = 0
        for chain in model.chains:
            if chain.chain_type != 'Ion':
                continue
            for residue in chain.get_residues():
                n_preserved += 1
                orig_name = residue.resname.strip().upper()
                if orig_name in ion_name_map:
                    new_name = ion_name_map[orig_name]
                    if residue.resname != new_name:
                        residue.resname = new_name
                        # Also rename the atom to match residue
                        for atom in residue.get_atoms():
                            atom.name = new_name
                            atom.fullname = f' {new_name:<3s}'
                        n_converted += 1

        if n_converted > 0:
            warnings.warn(
                f'Converted {n_converted} existing ions to CHARMM format',
                UserWarning
            )

        return n_preserved

    def create_water_box_coords(self) -> np.ndarray:
        """
        Creates a water box grid based on the chosen box type.

        Supported box types:
          - 'cube': Cubic grid with side length = box_dim
          - 'octa': Truncated octahedron (bounding cube = box_dim * sqrt(4/3))
          - 'rhdo': Rhombic dodecahedron (bounding cube = box_dim * sqrt(2))
          - 'ortho': Orthorhombic box with dimensions box_dims = (a, b, c)
          - 'tetra': Tetragonal box (a=b, c different)
          - 'hexa': Hexagonal prism

        Returns water molecules as array of shape (N_waters, 3, 3).
        """
        from crimm.Modeller.CrystalSDF import (
            sdf_cube, sdf_truncated_octahedron, sdf_rhombic_dodecahedron,
            sdf_orthorhombic, sdf_hexagonal, get_bounding_cube_size
        )

        # Determine grid dimensions based on crystal type
        if self.box_type == 'cube':
            grid_dims = self.box_dims
        elif self.box_type == 'octa':
            # Bounding cube for truncated octahedron
            grid_length = self.box_dim * math.sqrt(4 / 3)
            grid_dims = (grid_length, grid_length, grid_length)
        elif self.box_type == 'rhdo':
            # Bounding cube for rhombic dodecahedron
            grid_length = self.box_dim * math.sqrt(2)
            grid_dims = (grid_length, grid_length, grid_length)
        elif self.box_type in ('ortho', 'tetra'):
            grid_dims = self.box_dims
        elif self.box_type == 'hexa':
            # Hexagon fits in 2a x 2a square (approximately)
            a, _, c = self.box_dims
            grid_dims = (a * 1.2, a * 1.2, c)
        else:
            grid_dims = self.box_dims

        # Build water grid
        n_units = [int(math.ceil(d / BOXWIDTH)) for d in grid_dims]
        water_coords_expanded = self.water_unit_coords.reshape(-1, 3)

        # Generate grid points
        water_points = []
        for i in range(n_units[0]):
            for j in range(n_units[1]):
                for k in range(n_units[2]):
                    translation = np.array([i * BOXWIDTH, j * BOXWIDTH, k * BOXWIDTH])
                    for atom in water_coords_expanded:
                        water_points.append(atom + translation)

        water_points = np.array(water_points)

        # Recenter grid at origin
        translation_vec = -np.ptp(water_points, axis=0) / 2 - water_points.min(0)
        water_points += translation_vec

        # Reshape into water molecules (each with 3 atoms: O, H1, H2)
        water_box = water_points.reshape(-1, 3, 3)

        # For cubic box, return all waters within boundary (no SDF needed)
        if self.box_type == 'cube':
            # Filter waters within cubic boundary
            half = self.box_dim / 2
            selected_waters = []
            for water in water_box:
                oxygen = water[0]
                if all(abs(oxygen) <= half):
                    selected_waters.append(water)
            return np.array(selected_waters) if selected_waters else np.empty((0, 3, 3))

        # For other types, use SDF to filter waters
        selected_waters = []
        for water in water_box:
            oxygen = water[0]

            if self.box_type == 'octa':
                sdf_value = sdf_truncated_octahedron(oxygen, self.box_dim)
            elif self.box_type == 'rhdo':
                sdf_value = sdf_rhombic_dodecahedron(oxygen, self.box_dim)
            elif self.box_type in ('ortho', 'tetra'):
                sdf_value = sdf_orthorhombic(oxygen, self.box_dims)
            elif self.box_type == 'hexa':
                sdf_value = sdf_hexagonal(oxygen, self.box_dims[0], self.box_dims[2])
            else:
                # Fallback to cubic
                sdf_value = sdf_cube(oxygen, self.box_dim)

            # Select waters at least solvcut distance inside the boundary
            if sdf_value <= -self.solvcut:
                selected_waters.append(water)

        return np.array(selected_waters) if selected_waters else np.empty((0, 3, 3))

    def get_expelled_water_box_coords(self) -> np.ndarray:
        """
        Returns water molecules that are outside the solvcut distance from the solute.
        
        All crystal types (cube, octa, rhdo, ortho, tetra, hexa) are now filtered
        by their respective SDF in create_water_box_coords(), so only solvcut 
        filtering is needed here.
        """
        if self.box_type not in ("cube", "octa", "rhdo", "ortho", "tetra", "hexa"):
            raise ValueError(f"Unsupported box type: {self.box_type}")

        # create_water_box_coords() returns shape (N_waters, 3, 3) for all types
        water_box = self.create_water_box_coords()
        
        if len(water_box) == 0:
            return np.empty((0, 3, 3))
        
        # Use the oxygen atom (first atom) of each water for KDTree filtering.
        oxy_coords = np.array([water[0] for water in water_box])
        kd_tree = KDTree(self.coords)
        water_kd_tree = KDTree(oxy_coords)
        r = water_kd_tree.query_ball_tree(kd_tree, self.solvcut)
        cutoff_select = []
        for nei_list in r:
            cutoff_select.append(not bool(len(nei_list)))
        cutoff_select = np.array(cutoff_select)
        water_box = water_box[cutoff_select]
        return water_box

    def _create_new_water_chain(self, alphabet_index) -> Solvent:
        chain_id = 'W'+self.alphabet[alphabet_index]
        water_chain = Solvent(chain_id)
        water_chain.pdbx_description = 'water'
        water_chain.source = 'generated'
        return water_chain

    def _solvate_model(self):
        self.water_box_coords = self.get_expelled_water_box_coords()
        assert self.water_box_coords.shape[1:] == (3, 3), \
        f'Invalid water box coords shape {self.water_box_coords.shape}'

        # Create single water chain (PSF/CRD extended format supports large residue numbers)
        cur_water_chain = self._create_new_water_chain(0)
        water_chains = [cur_water_chain]
        segid = 'SOLV'

        for i, res_coords in enumerate(self.water_box_coords):
            resseq = i + 1  # Sequential residue numbering starting at 1

            water_res = Residue((' ', resseq, ' '), 'TIP3', segid)

            cur_oxygen = OH2.copy()
            cur_h1 = H1.copy()
            cur_h2 = H2.copy()

            OH2_coord, H1_coord, H2_coord = res_coords
            cur_oxygen.coord = OH2_coord
            cur_h1.coord = H1_coord
            cur_h2.coord = H2_coord

            water_res.add(cur_oxygen)
            water_res.add(cur_h1)
            water_res.add(cur_h2)
            cur_water_chain.add(water_res)

        for water_chain in water_chains:
            self.model.add(water_chain)
            if self._topo_loader is not None:
                self._topo_loader.generate_solvent(
                    water_chain, solvent_model='TIP3'
                )

        # Store box/crystal information for PSF/CRD title generation
        if not hasattr(self.model, '_solvation_info'):
            self.model._solvation_info = {}
        self.model._solvation_info['box_type'] = self.box_type
        self.model._solvation_info['box_dim'] = self.box_dim
        # Store box_dims for orthorhombic and related types
        if hasattr(self, 'box_dims') and self.box_dims is not None:
            self.model._solvation_info['box_dims'] = self.box_dims
        # Store crystal type info for output writers
        if self.box_type in CRYSTAL_TYPES:
            crystal_info = CRYSTAL_TYPES[self.box_type]
            self.model._solvation_info['charmm_name'] = crystal_info.charmm_name
            self.model._solvation_info['angles'] = crystal_info.angles

        return water_chains

    def _remove_generated_ions(self):
        """Removes any previously generated ion chains from the model."""
        remove_list = []
        for chain in self.model.chains:
            if chain.chain_type == 'Ion' and chain.source == 'generated':
                remove_list.append(chain.id)
        for chain_id in remove_list:
            self.model.detach_child(chain_id)
        if len(remove_list) > 0:
            warnings.warn(
                f'Removed {len(remove_list)} previously generated ion chains',
                UserWarning
            )
            
    def add_balancing_ions(
            self, present_charge = None, cation='SOD', anion='CLA', skip_undefined=True,
            remove_generated_ions=True
        ) -> Ion:
        """DEPRECATED: Use add_ions(concentration=0.0) instead.

        Add balancing ions to the solvated entity to bring total charge to zero.
        The default cation is Na+ and the default anion is Cl-. If the entity is
        not a solvated entity, a ValueError will be raised. A random selection of
        water molecules in the water box will be replaced with balancing ions.
        Returns a chain containing the balancing ions.
        
        Parameters
        ----------
        entity : Structure, Model, or Chain level entity
            The solvated entity to add balancing ions to.
        present_charge : int, optional
            The present charge of the solvated entity. If None, the charge will be
            calculated from the entity. The default is None. If for any reason you
            want to balance the charge to a non-zero value, you can specify it here.
        cation : str, optional
            The cation to use. The default is 'SOD' (Na+).
        anion : str, optional
            The anion to use. The default is 'CLA' (Cl-).
        skip_undefined : bool, optional
            If True, chains without defined topology will be assumed to have zero
            charge. If False, a ValueError will be raised if any chain has undefined
            charge. The default is True.
        remove_generated_ions : bool, optional
            If True, any previously generated balancing ion chains will be removed
            before adding new ions. The default is True.
                
        Returns
        -------
        ion_chain : Chain
            A chain containing the balancing ions.
        """
        warnings.warn(
            "add_balancing_ions() is deprecated. Use add_ions(concentration=0.0) instead.",
            DeprecationWarning,
            stacklevel=2
        )
        if remove_generated_ions:
            self._remove_generated_ions()
        solvents = [chain for chain in self.model if chain.chain_type == 'Solvent']
        if len(solvents) == 0:
            raise ValueError(
                'Entity must be a solvated Structure or Model'
            )
        
        if present_charge is None:
            charge_dict = {}
            total_charges = 0
            for chain in self.model:
                if chain.chain_type == 'Solvent':
                    continue
                if chain.total_charge is None:
                    if not skip_undefined:
                        raise ValueError(
                            f'Chain {chain.id} has no topology definition for atom charge! '
                            'Cannot calculate total charge.'
                        )

                    warnings.warn(
                        f'Chain {chain.id} has no topology definition for atom charge! '
                        'Assume zero charge.',
                    )
                charge_dict[chain.id] = chain.total_charge
                total_charges+=chain.total_charge
            print(f'Total charges before adding ions: {total_charges}')
            for chain_id, charge in charge_dict.items():
                print(f'  [Chain {chain_id}] {charge}')
        else:
            total_charges = present_charge
            print(f'Total charges before adding ions: {total_charges}')

        if total_charges == 0:
            warnings.warn('No balancing ions needed', UserWarning)
            return None
        
        if abs(int(total_charges)-total_charges) > 1e-2:
            raise ValueError(
                f'Invalid total charge {total_charges} for balancing ions! '
                'Total charge must be an integer.'
            )

        if total_charges > 0:
            ion_list = [anion for i in range(int(total_charges))]
        else:
            ion_list = [cation for i in range(int(-total_charges))]

        new_ion_chain = self._create_ion_chain(solvents, ion_list)
        self.model.add(new_ion_chain)
        if self._topo_loader is not None:
            self._topo_loader.generate(new_ion_chain)

        return new_ion_chain
        
    def _create_ion_chain(self, solvents, ion_list):
        
        water_res = [res for chain in solvents for res in chain]
        chosen_waters = choices(water_res, k=len(ion_list))

        new_ion_chain = Ion('IA')
        new_ion_chain.source = 'generated'
        ion_names = ', '.join(set(ion_list))
        new_ion_chain.pdbx_description = f"balancing ions ({ion_names})"
        for i, (chosen_water, ion_name) in enumerate(zip(chosen_waters, ion_list), start=1):
            if 'OH2' in chosen_water:
                oxy_coord = chosen_water['OH2'].coord
            elif 'O' in chosen_water:
                oxy_coord = chosen_water['O'].coord
            else:
                raise KeyError(f'No oxygen atom present in water {chosen_water}')
            if ion_name not in _topo_def.res_defs:
                raise ValueError(
                    f'Ion {ion_name} not exist in water_ions.rtf. Ion names must be '
                    f'in {CHARMM_PDB_ION_NAMES.keys()}'
                )
            ion_res = _topo_def.res_defs[ion_name].create_residue(resseq=i)
            ion_res.atoms[0].coord = oxy_coord
            new_ion_chain.add(ion_res)
            if (water_chain:=chosen_water.parent) is not None:
                water_chain.detach_child(chosen_water.id)

        return new_ion_chain

    # =========================================================================
    # New Ion Calculation Methods (SPLIT, SLTCAP, Add-Neutralize)
    # =========================================================================
    # TODO: Move this inside of the `get_charge` function in Utils
    def _get_ion_chain_charge(self, chain) -> float:
        """Calculate total charge of an ion chain from known ion valences.

        Since ion chains typically don't have topology, we calculate their
        charge from the ION_VALENCES table based on residue names.

        Parameters
        ----------
        chain : Ion
            An ion chain.

        Returns
        -------
        float
            Total charge of all ions in the chain.
        """
        total = 0.0
        for res in chain.get_residues():
            name = res.resname.strip().upper()
            if name in ION_VALENCES:
                total += ION_VALENCES[name]
            else:
                warnings.warn(
                    f"Unknown ion '{name}' in chain {chain.id}, assuming charge 0. "
                    f"Add to ION_VALENCES if needed.",
                    UserWarning
                )
        return total

    def _calculate_system_charge(self, skip_undefined: bool = True) -> float:
        """Calculate total system charge from all non-solvent chains.

        Includes:
        - Protein/nucleic acid chains (from topology)
        - Existing ion chains (from ION_VALENCES table)

        Parameters
        ----------
        skip_undefined : bool
            If True, assume zero charge for chains without topology.
            If False, raise ValueError.

        Returns
        -------
        float
            Total system charge in electron units.
        """
        total_charge = 0.0
        for chain in self.model:
            if chain.chain_type == 'Solvent':
                continue
            if chain.chain_type == 'Ion':
                # Include existing ion charges (e.g., Zn2+, Mg2+, Ca2+)
                ion_charge = self._get_ion_chain_charge(chain)
                total_charge += ion_charge
                continue
            if chain.total_charge is None:
                if not skip_undefined:
                    raise ValueError(
                        f'Chain {chain.id} has no topology definition for atom charge! '
                        'Cannot calculate total charge.'
                    )
                warnings.warn(
                    f'Chain {chain.id} has no topology definition for atom charge! '
                    'Assuming zero charge.',
                    UserWarning
                )
                continue
            total_charge += chain.total_charge
        return total_charge

    def _count_waters(self) -> int:
        """Count total number of water molecules in solvent chains."""
        n_water = 0
        for chain in self.model:
            if chain.chain_type == 'Solvent':
                n_water += len(list(chain.get_residues()))
        return n_water

    def _calculate_solvent_volume(self) -> float:
        """Calculate solvent volume in Å³ based on box geometry.

        Returns
        -------
        float
            Solvent volume in Å³.
        """
        if self.box_dim is None:
            raise ValueError("Box dimensions not set. Run solvate() first.")

        volume_factor = BOX_VOLUME_FACTORS.get(self.box_type, 1.0)
        return volume_factor * (self.box_dim ** 3)

    def _calculate_n0(self, n_water: int, concentration: float) -> float:
        """Calculate N₀ (number of neutral ion pairs) for target concentration.

        Parameters
        ----------
        n_water : int
            Number of water molecules.
        concentration : float
            Target salt concentration in Molar.

        Returns
        -------
        float
            Number of neutral ion pairs (may be fractional).
        """
        return n_water * concentration / WATER_MOLARITY

    def _select_ion_method(
        self, n_water: int, charge: float, concentration: float
    ) -> tuple:
        """Auto-select best ion calculation method based on N₀/|Q| ratio.

        Parameters
        ----------
        n_water : int
            Number of water molecules.
        charge : float
            System charge in electron units.
        concentration : float
            Target salt concentration in Molar.

        Returns
        -------
        tuple
            (method_name, explanation_string)
        """
        n0 = self._calculate_n0(n_water, concentration)

        if charge == 0:
            return 'split', "Neutral system - all methods equivalent"

        if n0 == 0:
            return 'add_neutralize', "No salt concentration - neutralization only"

        ratio = n0 / abs(charge)

        if ratio >= 2.0:
            return 'split', f"N₀/|Q| = {ratio:.1f} ≥ 2: SPLIT excellent (<1% error)"
        elif ratio >= 1.0:
            return 'split', f"N₀/|Q| = {ratio:.1f} ≥ 1: SPLIT acceptable (~7% error)"
        elif ratio >= 0.5:
            return 'sltcap', f"N₀/|Q| = {ratio:.2f} < 1: Using SLTCAP (SPLIT error ~17%)"
        else:
            return 'sltcap', f"N₀/|Q| = {ratio:.2f} << 1: SLTCAP required (high charge)"

    def _adjust_for_neutralization(
        self, n_cation: float, n_anion: float, charge: float
    ) -> tuple:
        """Adjust rounded ion counts to ensure exact charge neutralization.

        Parameters
        ----------
        n_cation : float
            Calculated cation count (may be fractional).
        n_anion : float
            Calculated anion count (may be fractional).
        charge : float
            System charge in electron units.

        Returns
        -------
        tuple
            (n_cation, n_anion) as integers, adjusted for exact neutralization.
        """
        # Ensure non-negative and round
        n_cation = max(0, round(n_cation))
        n_anion = max(0, round(n_anion))

        # Adjust for exact neutralization
        # Net ion charge = n_cation - n_anion (assumes monovalent ions)
        # For neutralization: n_cation - n_anion = -charge
        net_ion_charge = n_cation - n_anion
        target_ion_charge = -int(round(charge))

        if net_ion_charge != target_ion_charge:
            diff = target_ion_charge - net_ion_charge
            if diff > 0:
                n_cation += 1
            else:
                n_anion += 1

        return int(n_cation), int(n_anion)

    def _ions_split(
        self, charge: float, n_water: int, concentration: float
    ) -> tuple:
        """Calculate ion counts using SPLIT method.

        SPLIT: N± = N₀ ∓ Q/2

        Reference: Machado & Pantano (2020) J. Chem. Theory Comput. 16:1367-1372

        Parameters
        ----------
        charge : float
            System charge in electron units.
        n_water : int
            Number of water molecules.
        concentration : float
            Target salt concentration in Molar.

        Returns
        -------
        tuple
            (n_cation, n_anion) as integers.
        """
        n0 = self._calculate_n0(n_water, concentration)
        n_cation = n0 - charge / 2.0
        n_anion = n0 + charge / 2.0
        return self._adjust_for_neutralization(n_cation, n_anion, charge)

    def _ions_sltcap(
        self, charge: float, n_water: int, concentration: float
    ) -> tuple:
        """Calculate ion counts using SLTCAP method.

        SLTCAP: N± = N₀ × [√(1 + (Q/(2N₀))²) ∓ Q/(2N₀)]

        Reference: Schmit et al. (2018) J. Chem. Theory Comput. 14:1823-1827

        Parameters
        ----------
        charge : float
            System charge in electron units.
        n_water : int
            Number of water molecules.
        concentration : float
            Target salt concentration in Molar.

        Returns
        -------
        tuple
            (n_cation, n_anion) as integers.
        """
        n0 = self._calculate_n0(n_water, concentration)

        if n0 == 0:
            # No salt - just neutralization
            if charge > 0:
                return 0, int(abs(charge))
            elif charge < 0:
                return int(abs(charge)), 0
            else:
                return 0, 0

        ratio = charge / (2 * n0)
        sqrt_term = math.sqrt(1 + ratio ** 2)

        n_cation = n0 * (sqrt_term - ratio)
        n_anion = n0 * (sqrt_term + ratio)

        return self._adjust_for_neutralization(n_cation, n_anion, charge)

    def _ions_add_neutralize(
        self, charge: float, n_water: int, concentration: float
    ) -> tuple:
        """Calculate ion counts using Add-Neutralize (AN) method.

        AN method: Add N₀ pairs, then add |Q| counterions.

        WARNING: This method overestimates effective salt concentration
        for charged systems. Use SPLIT or SLTCAP for better accuracy.

        Parameters
        ----------
        charge : float
            System charge in electron units.
        n_water : int
            Number of water molecules.
        concentration : float
            Target salt concentration in Molar.

        Returns
        -------
        tuple
            (n_cation, n_anion) as integers.
        """
        n0 = self._calculate_n0(n_water, concentration)
        n0_int = round(n0)

        if charge > 0:
            n_cation = n0_int
            n_anion = n0_int + int(abs(charge))
        elif charge < 0:
            n_cation = n0_int + int(abs(charge))
            n_anion = n0_int
        else:
            n_cation = n0_int
            n_anion = n0_int

        return int(n_cation), int(n_anion)

    def _place_ions_monte_carlo(
        self,
        solvents: list,
        ion_list: list,
        min_dist_solute: float = 5.0,
        min_dist_ion: float = 5.0,
        max_attempts: int = 1000
    ) -> list:
        """Place ions using Monte-Carlo with distance constraints.

        Parameters
        ----------
        solvents : list
            List of Solvent chain objects.
        ion_list : list
            List of ion names to place (e.g., ['SOD', 'SOD', 'CLA']).
        min_dist_solute : float
            Minimum distance from solute atoms in Å.
        min_dist_ion : float
            Minimum distance between placed ions in Å.
        max_attempts : int
            Maximum attempts to place each ion before giving up.

        Returns
        -------
        list
            List of (water_residue, ion_name) tuples for successful placements.
        """
        # Get all water residues and their oxygen coordinates
        water_residues = []
        water_coords = []
        for chain in solvents:
            for res in chain.get_residues():
                if 'OH2' in res:
                    water_residues.append(res)
                    water_coords.append(res['OH2'].coord)
                elif 'O' in res:
                    water_residues.append(res)
                    water_coords.append(res['O'].coord)

        if len(water_residues) == 0:
            raise ValueError("No water molecules found for ion placement")

        water_coords = np.array(water_coords)

        # Build KDTree of solute atoms
        solute_coords = []
        for chain in self.model:
            if chain.chain_type not in ('Solvent', 'Ion'):
                for atom in chain.get_atoms():
                    solute_coords.append(atom.coord)

        if len(solute_coords) > 0:
            solute_coords = np.array(solute_coords)
            solute_tree = KDTree(solute_coords)
        else:
            solute_tree = None

        # Track placed ion positions and used water indices
        placed_ion_coords = []
        used_water_indices = set()
        placements = []

        for ion_name in ion_list:
            placed = False
            attempts = 0

            # Shuffle water indices for random selection
            available_indices = [
                i for i in range(len(water_residues))
                if i not in used_water_indices
            ]
            random.shuffle(available_indices)

            for idx in available_indices:
                if attempts >= max_attempts:
                    break
                attempts += 1

                coord = water_coords[idx]

                # Check distance from solute
                if solute_tree is not None:
                    dist_to_solute = solute_tree.query(coord)[0]
                    if dist_to_solute < min_dist_solute:
                        continue

                # Check distance from other placed ions
                if len(placed_ion_coords) > 0:
                    ion_coords_array = np.array(placed_ion_coords)
                    dists = np.linalg.norm(ion_coords_array - coord, axis=1)
                    if np.min(dists) < min_dist_ion:
                        continue

                # Accept this placement
                placements.append((water_residues[idx], ion_name))
                placed_ion_coords.append(coord)
                used_water_indices.add(idx)
                placed = True
                break

            if not placed:
                warnings.warn(
                    f"Could not place ion {ion_name} after {max_attempts} attempts. "
                    f"Consider reducing min_dist_solute or min_dist_ion.",
                    UserWarning
                )

        return placements

    def add_ions(
        self,
        concentration: float = 0.15,
        method: str = 'auto',
        cation: str = 'SOD',
        anion: str = 'CLA',
        min_dist_solute: float = 5.0,
        min_dist_ion: float = 5.0,
        skip_undefined: bool = True,
        remove_generated_ions: bool = True,
    ) -> Optional[Ion]:
        """Add ions to achieve target salt concentration.

        Supports three calculation methods:
        - 'split': SPLIT method (Machado & Pantano, 2020) - good for N₀/|Q| ≥ 1
        - 'sltcap': SLTCAP method (Schmit et al., 2018) - accurate for any system
        - 'add_neutralize': Simple add-neutralize (may overestimate concentration)
        - 'auto': Automatically select best method based on N₀/|Q| ratio

        Parameters
        ----------
        concentration : float
            Target salt concentration in Molar. Default 0.15 M (150 mM).
            Use concentration=0 to only neutralize without adding salt.
        method : str
            Ion calculation method: 'auto', 'split', 'sltcap', or 'add_neutralize'.
        cation : str
            CHARMM ion name for cation. Default 'SOD' (Na+).
            Note: Only monovalent ions (SOD, POT, LIT, etc.) are supported.
        anion : str
            CHARMM ion name for anion. Default 'CLA' (Cl-).
            Note: Only monovalent ions (CLA) are supported.
        min_dist_solute : float
            Minimum distance from solute for ion placement in Å. Default 5.0.
        min_dist_ion : float
            Minimum distance between ions in Å. Default 5.0.
        skip_undefined : bool
            If True, assume zero charge for chains without topology. Default True.
        remove_generated_ions : bool
            If True, any previously generated ion chains will be removed. Default True.

        Returns
        -------
        Ion or None
            Chain containing the added ions, or None if no ions needed.

        References
        ----------
        - Schmit et al. (2018) J. Chem. Theory Comput. 14:1823-1827 (SLTCAP)
        - Machado & Pantano (2020) J. Chem. Theory Comput. 16:1367-1372 (SPLIT)
        """
        # Remove previously generated ions
        if remove_generated_ions:
            self._remove_generated_ions()

        # Verify system is solvated
        solvents = [c for c in self.model if c.chain_type == 'Solvent']
        if len(solvents) == 0:
            raise ValueError(
                "Entity must be solvated before adding ions. Run solvate() first."
            )

        # Calculate system properties
        charge = self._calculate_system_charge(skip_undefined)
        n_water = self._count_waters()

        # Select method if auto
        if method == 'auto':
            method, method_reason = self._select_ion_method(n_water, charge, concentration)
        else:
            method_reason = f"User-selected method: {method.upper()}"

        # Calculate ion counts
        if method == 'split':
            n_cation, n_anion = self._ions_split(charge, n_water, concentration)
        elif method == 'sltcap':
            n_cation, n_anion = self._ions_sltcap(charge, n_water, concentration)
        elif method == 'add_neutralize':
            n_cation, n_anion = self._ions_add_neutralize(charge, n_water, concentration)
            # Warn about AN method limitations
            if charge != 0 and concentration > 0:
                n0 = self._calculate_n0(n_water, concentration)
                c_eff = concentration * math.sqrt(1 + abs(charge) / (n0 + 1e-10))
                warnings.warn(
                    f"Add-Neutralize method may overestimate effective salt concentration.\n"
                    f"For system charge Q={charge:.0f}, effective concentration ≈ {c_eff*1000:.0f} mM "
                    f"vs target {concentration*1000:.0f} mM.\n"
                    f"Consider using method='split' or method='sltcap' for better accuracy.",
                    UserWarning
                )
        else:
            raise ValueError(f"Unknown method: {method}. Use 'auto', 'split', 'sltcap', or 'add_neutralize'.")

        # Build ion list
        ion_list = [cation] * n_cation + [anion] * n_anion

        if len(ion_list) == 0:
            print("No ions needed for this system.")
            return None

        # Calculate N₀ for reporting
        n0 = self._calculate_n0(n_water, concentration)
        ratio = n0 / abs(charge) if charge != 0 else float('inf')

        # Print informative message
        print(f"\nIon calculation using {method.upper()} method:")
        print(f"  {method_reason}")
        print(f"  System charge: {charge:+.0f} e")
        print(f"  Water molecules: {n_water:,}")
        print(f"  Target concentration: {concentration*1000:.0f} mM")
        print(f"  N₀ (neutral pairs): {n0:.1f}")
        if charge != 0:
            print(f"  N₀/|Q| ratio: {ratio:.2f}")
        print(f"\n  Adding ions:")
        print(f"    {cation}: {n_cation}")
        print(f"    {anion}: {n_anion}")

        # Place ions using Monte-Carlo
        placements = self._place_ions_monte_carlo(
            solvents, ion_list, min_dist_solute, min_dist_ion
        )

        if len(placements) == 0:
            warnings.warn("No ions could be placed!", UserWarning)
            return None

        # Create ion chain
        new_ion_chain = Ion('IA')
        new_ion_chain.source = 'generated'
        segid = 'IONS'
        ion_names_str = ', '.join(sorted(set(ion_list)))
        new_ion_chain.pdbx_description = f"ions ({ion_names_str}) at {concentration*1000:.0f} mM"

        for i, (water_res, ion_name) in enumerate(placements, start=1):
            if 'OH2' in water_res:
                oxy_coord = water_res['OH2'].coord
            else:
                oxy_coord = water_res['O'].coord

            ion_res = _topo_def.res_defs[ion_name].create_residue(resseq=i)
            ion_res.atoms[0].coord = oxy_coord
            ion_res.segid = segid
            new_ion_chain.add(ion_res)

            # Remove water
            if (water_chain := water_res.parent) is not None:
                water_chain.detach_child(water_res.id)

        self.model.add(new_ion_chain)
        if self._topo_loader is not None:
            self._topo_loader.generate(new_ion_chain)

        # Report final charge
        ion_charge = n_cation - n_anion
        final_charge = charge + ion_charge
        print(f"\n  Final system charge: {final_charge:+.0f} e")
        if abs(final_charge) > 0.5:
            warnings.warn(
                f"System is not fully neutralized (charge = {final_charge:+.0f} e). "
                f"This may cause issues with PME.",
                UserWarning
            )

        # Store solvation metadata for PSF/CRD title generation
        if not hasattr(self.model, '_solvation_info'):
            self.model._solvation_info = {}
        self.model._solvation_info['concentration'] = concentration
        self.model._solvation_info['n_cation'] = n_cation
        self.model._solvation_info['n_anion'] = n_anion
        self.model._solvation_info['cation'] = cation
        self.model._solvation_info['anion'] = anion
        self.model._solvation_info['method'] = method

        return new_ion_chain





"""The structure class, representing a macromolecular structure."""
import warnings
from Bio.PDB.Structure import Structure as _Structure

class Structure(_Structure):
    """The extended Structure class contains a collection of Model instances.
    Derived from Biopython's Bio.PDB.Model and compatible with Biopython functions
    """

    def __init__(self, id) -> None:
        super().__init__(id)
        self.pdb_id = None
        self._header = None
        self.resolution = None
        self.method = None
        self.assemblies = None
        self.cell_info = None

    @property
    def header(self):
        """Return the header information of the structure."""
        return self._header
    
    @header.setter
    def header(self, header_dict):
        """Set the header information of the structure."""
        self._header = header_dict
        if ('idcode' in header_dict) and (
            header_dict['idcode'] is not None
            ) and ('title' in header_dict['idcode']):
            for model in self.get_models():
                model.pdbx_description = header_dict['idcode']['title']

    def set_pdb_id(self, pdb_id):
        """Set the PDB ID of this structure."""
        if self.pdb_id is not None:
            warnings.warn(
                f"Overwriting PDB ID {self.pdb_id} with {pdb_id}"
            )
        self.pdb_id = pdb_id
        # we set pdb_id on models to make many other routines easier
        for model in self.get_models():
            model.set_pdb_id(pdb_id)

    def __repr__(self):
        return f"<Structure id={self.get_id()} Models={len(self)}>"

    def expanded_view(self):
        """Print the hierarchy tree of this model."""
        hierarchy_str = repr(self)
        if len(self) == 0:
            return hierarchy_str
        first_model = self.child_list[0]
        hierarchy_str+='\n│\n├───'+first_model.expanded_view()
        if len(self) > 1:
            hierarchy_str+=f'\n[{len(self)-1} models truncated ...]'
        return hierarchy_str
    
    def _ipython_display_(self):
        """Return the nglview interactive visualization window"""
        if len(self) == 0:
            return
        from crimm.Visualization import show_nglview_multiple
        from IPython.display import display
        # Only show the first model for now
        display(show_nglview_multiple(self.child_list[0].child_list))
        print(self.expanded_view())

    @property
    def models(self):
        """Alias for child_list. Returns the list of models in this structure."""
        return self.child_list

    def get_atoms(self, include_alt=False):
        """Return a generator of all atoms from this structure. If include_alt is True, the 
        disordered residues will be expanded and altloc of disordered atoms will be included."""
        for model in self:
            yield from model.get_atoms(include_alt=include_alt)

    def reset_atom_serial_numbers(self, include_alt=True):
        """Reset all atom serial numbers in the structure starting from 1."""
        for model in self:
            model.reset_atom_serial_numbers(include_alt=include_alt)


## for charmm pdb files
nucleic_letters_1to3 = {
    'A': 'ADE', 'C': 'CYT', 'G': 'GUA', 'T': 'THY', 'U': 'URA',
}

NUCLEOSIDE_PHOS = ( # Nucleoside phosphates and phosphonates
    'ATP', 'ADP', 'AMP', 'ANP', 'ACP', 
    'CTP', 'CDP', 'CMP',
    'GCP', 'GTP', 'GDP', 'GNP',
    'UTP', 'UDP', 'UMP', 'UNP', 'UCP', 
    'TMP', 'TTP',
    'IMP', 'IDP'
)
# GlYCOSYLATION = ('NAG', 'NDG', 'NGA')
# MONOSACCGARIDES = (
#     'GLC', 'GAL', 'FRU', '3MK', '4GL', '4N2', 'AFD',
#     'ALL', 'ALT', 'ARA', 'BGC', 'BMA', 'BXY', 'CEL',
#     'DGL', 'FUC', 'GIV', 'GL0', 'GLA', 'GUP', 'BDF',
#     'GXL', 'MAN', 'SDY', 'SHD', 'WOO', 'Z0F', 'Z2D',
#     'Z6H', 'Z8H', 'Z8T', 'ZCD', 'ZEE', 
# )
glucopyranose = 'C(C1C(C(C(C(O1)O)O)O)O)O'
amino_deoxy_glucopyranose = 'NC1C(C(C(OC1O)CO)O)O'
fructofuranose = 'C(C1C(C(C(O1)(CO)O)O)O)O'
fructopyranose = 'C1C(C(C(C(O1)(CO)O)O)O)O'

ALKALI_METALS = ('Li','Na','K','Rb','Cs','Fr')
ALKALI_EARTH_METALS = ('Be','Mg','Ca','Sr','Ba','Ra')
TRANSITION_METALS = (
    'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
    'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd',
    'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', 'Hg',
    'Rf', 'Db', 'Sg', 'Bh', 'Hs'
)
ALL_METALS = tuple((*ALKALI_METALS, *ALKALI_EARTH_METALS, *TRANSITION_METALS))

PDB_CHARMM_ION_NAMES = {
    "LI": "LIT",
    "NA": "SOD",
    "MG": "MG",
    "K": "POT",
    "CA": "CAL",
    "RB": "RUB",
    "CS": "CES",
    "BA": "BAR",
    "ZN": "ZN", # However, ZN2 is the resname used in CHARMM
    "CD": "CAD",
    "CL": "CLA",
}

CHARMM_PDB_ION_NAMES = {k:v for v,k in PDB_CHARMM_ION_NAMES.items()}



# Copyright (C) 2023, Truman Xu, Brooks Lab at the University of Michigan
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.

"""Module containing the parser class for constructing structures from mmCIF 
files from PDB"""
from copy import deepcopy
import warnings
import numpy as np
from Bio.PDB.PDBExceptions import PDBConstructionWarning
from crimm.StructEntities.Atom import Atom
from crimm.IO.MMCIF2Dict import MMCIF2Dict
from crimm.IO.StructureBuilder import StructureBuilder
from crimm.StructEntities.Chain import (
    Chain, PolymerChain, Heterogens, Oligosaccharide, Solvent, Macrolide
)
from crimm.StructEntities.Model import Model
from crimm.Utils.StructureUtils import (
    get_coords, index_to_letters, letters_to_index
)

class MMCIFParser:
    """Parser class for standard mmCIF files from PDB"""
    def __init__(
            self,
            first_model_only = True,
            use_bio_assembly = True,
            include_solvent = True,
            include_hydrogens = False,
            strict_parser = True,
            QUIET = False
        ):
        self._structure_builder = StructureBuilder()
        self.QUIET = QUIET
        self.first_model_only = first_model_only
        self.use_bio_assembly = use_bio_assembly
        self.include_solvent = include_solvent
        self.include_hydrogens = include_hydrogens
        self.strict_parser = strict_parser
        self.cifdict = None
        self.model_template = None
        self.structure_id = None
        self.assembly_dict = None
        self.symmetry_ops = None
        self.skipped_chain_id = []

    @staticmethod
    def _cif_find_resolution(cifdict):
        """Find structure resolution information from the parsed mmCIF dictionary"""
        resolution_keys = [
            ("refine", "ls_d_res_high"),
            ("refine_hist", "d_res_high"),
            ("em_3d_reconstruction", "resolution"),
        ]
        for key, subkey in resolution_keys:
            resolution = cifdict.level_two_get(key, subkey)
            if resolution is not None and (value:=resolution[0]) is not None:
                return value

    @staticmethod
    def _cif_get_header(cifdict):
        """Get header information from the parsed mmCIF dictionary"""
        date =  cifdict.level_two_get(
            "pdbx_database_status", "recvd_initial_deposition_date"
        )
        if date is not None:
            date = date[0]
        else:
            date = None
        header = {
            "name": cifdict.get("data"),
            "keywords": cifdict.retrieve_single_value_dict("struct_keywords"),
            "citation": cifdict.get("citation"),
            "idcode": cifdict.retrieve_single_value_dict('struct'),
            "deposition_date": date,
        }

        return header

    def get_structure(self, filepath, structure_id = None):
        """Return the structure.

        Arguments:
         :structure_id: string, the id that will be used for the structure
         :filepath: path to mmCIF file, OR an open text mode file handle

        """
        with warnings.catch_warnings():
            if self.QUIET:
                warnings.filterwarnings("ignore", category=PDBConstructionWarning)
            # mmCIF will be parsed into dictionary first and then namedtuples
            # to gather all the necessary info to construct the structure
            self.cifdict = MMCIF2Dict(filepath)
            self.structure_id = structure_id
            self.model_template = self.create_model_template()
            # get bioassembly info
            self.assembly_dict = self.create_assembly_dict()
            # find crystal symmetry operation
            self.symmetry_ops = self._cif_find_symmetry_info()
            self._build_structure(structure_id)
            # set additional info on the structure
            structure_method = None
            struct_method_list = self.cifdict.level_two_get("exptl", "method")
            if struct_method_list is not None:
                structure_method = struct_method_list[0]
                
            self._structure_builder.set_structure_method(structure_method)
            resolution = self._cif_find_resolution(self.cifdict)
            self._structure_builder.set_resolution(resolution)
            header = self._cif_get_header(self.cifdict)
            self._structure_builder.set_header(header)

        return self._structure_builder.get_structure()

    def create_polymer_chain_dict(self):
        """Create a dictionary for the PolymerChain object from mmCIF. Empty 
        PolymerChain classes will be created, and information on entity id, 
        author chain id, reported sequence, and reported missing residues will 
        be assigned. Return a dictionary with empty PolymerChain objects keyed
        by entity id (int)"""
        zero_occupancy_residues = self.cifdict.create_namedtuples(
            'pdbx_unobs_or_zero_occ_residues'
        )
        missing_res_dict = dict()
        for res in zero_occupancy_residues:
            if res.auth_asym_id not in missing_res_dict:
                missing_res_dict[res.auth_asym_id] = []
            missing_res_dict[res.auth_asym_id].append(
                (res.label_seq_id, res.label_comp_id)
            )

        entity_poly = self.cifdict.create_namedtuples('entity_poly')
        entity_poly_dict = dict()
        for entity in entity_poly:
            auth_chain_ids = str(entity.pdbx_strand_id).split(',')
            for chain_id in auth_chain_ids:
                entity_poly_dict[chain_id] = entity

        entity_poly_seq = self.cifdict.create_namedtuples('entity_poly_seq')
        reported_res_dict = dict()
        for res in entity_poly_seq:
            if res.entity_id not in reported_res_dict:
                reported_res_dict[res.entity_id] = []
            reported_res_dict[res.entity_id].append((res.num, res.mon_id))

        polymer_dict = dict()
        for auth_chain_id, entity_info in entity_poly_dict.items():
            entity_id = entity_info.entity_id
            chain = PolymerChain(
                None,
                entity_id,
                auth_chain_id,
                entity_info.type,
                entity_info.pdbx_seq_one_letter_code,
                entity_info.pdbx_seq_one_letter_code_can,
                reported_res_dict[entity_info.entity_id],
                missing_res_dict.get(auth_chain_id)
            )
            polymer_dict[entity_id] = chain
        return polymer_dict

    def _set_chain_attr(self, entity, chain):
        """Set the info from mmcif as attribute for the chain"""
        for k, v in entity._asdict().items():
            if k != 'id' and k != 'type' and v is not None:
                setattr(chain, k, v)

    def create_model_template(self):
        """Set up empty templates for the chains in a template model. The chain 
        types are determined based on the entity type in mmCIF, and a model object
        containing the empty chains will be returned."""
        model_temp = Model('template')

        poly_dict = self.create_polymer_chain_dict()
        entity_info = self.cifdict.create_namedtuples('entity')

        for entity in entity_info:
            if entity.type == 'polymer':
                cur_chain = poly_dict[entity.id]
            elif entity.type == 'non-polymer':
                cur_chain = Heterogens(entity.id)
            elif entity.type == 'water':
                cur_chain = Solvent(entity.id)
            elif entity.type == 'branched':
                cur_chain = Oligosaccharide(entity.id)
            elif entity.type == 'macrolide':
                cur_chain = Macrolide(entity.id)
            elif not self.strict_parser:
                cur_chain = Chain(entity.id)
            else:
                raise TypeError(f'Unknown Chain Type: {entity.type}')
            self._set_chain_attr(entity, cur_chain)
            model_temp.add(cur_chain)
        return model_temp

    def create_atom_from_namedtuple(self, atom_entry):
        """Create an Atom object from the namedtuple of an "atom_site" entry
        from the mmCIF dict

        Arguments:
         :param atom_entry: A namedtuple created from mmCIF "atom_site" entries
                           with field names as attributes in the tuple
         :type atom_entry: namedtuple
        """
        coord = np.array(
                [atom_entry.Cartn_x, atom_entry.Cartn_y, atom_entry.Cartn_z]
        )
        altloc = atom_entry.label_alt_id
        if altloc is None:
            altloc = ' '
        else:
            altloc = str(atom_entry.label_alt_id)

        atom = Atom(
            name = atom_entry.label_atom_id,
            fullname = atom_entry.label_atom_id,
            coord = coord,
            bfactor = atom_entry.B_iso_or_equiv,
            occupancy = atom_entry.occupancy,
            altloc = altloc,
            serial_number = atom_entry.id,
            element = atom_entry.type_symbol,
        )
        return atom

    @staticmethod
    def _assign_hetflag(fieldname, resname):
        if fieldname != "HETATM":
            return ' '
        if resname in ("HOH", "WAT"):
            return "W"
        return "H"

    def _create_residue_dict_entry(self, atom_entry, resseq):
        """Create a residue dictionary entry with empty atom list"""
        resname = str(atom_entry.label_comp_id)
        hetatm_flag = self._assign_hetflag(
            atom_entry.group_PDB, resname
        )
        icode = atom_entry.pdbx_PDB_ins_code
        if icode is None:
            icode = ' '
        res_id = (hetatm_flag, resseq, icode)
        res_dict_entry = {
            "resname": resname,
            "res_id": res_id,
            "author_seq_id": atom_entry.auth_seq_id,
            "atom_list": []
        }
        return res_dict_entry
        
    def create_atom_site_entry_dict(self):
        """Create a dictionary containing structured data from all "atom_site" 
        fields in mmCIF. Return a dictionary that contains four levels, which 
        are keyed by [model_id,[chain_id,[resseq]]].

        The dictionary adopt the strucuture as following.
        structure_dict : {
            model_dict[model_id] : {
                chain_dict[chain_id] : {
                    res_dict[resseq] : {
                        "resname": str,
                        "res_id": Tuple[hetflag: str, resseq: int, icode: str],
                        "atom_list": List[atoms: Atom]
                    }
                }
            }
        }
        """
        atom_site = self.cifdict.create_namedtuples('atom_site')
        model_dict = dict()

        for atom_entry in atom_site:
            if not self.include_hydrogens and atom_entry.type_symbol == 'H':
                continue
            model_num = atom_entry.pdbx_PDB_model_num
            if model_num not in model_dict:
                if self.first_model_only and len(model_dict) > 0:
                    break
                model_dict[model_num] = dict()
            entity_dict = model_dict[model_num]
            entity_id = atom_entry.label_entity_id
            if entity_id not in entity_dict:
                entity_dict[entity_id] = dict()
            chain_dict = entity_dict[entity_id]
            chain_id = atom_entry.label_asym_id
            if chain_id not in chain_dict:
                last_auth_seq = None
                het_chain_resseq = 0
                chain_dict[chain_id] = dict()
            res_dict = chain_dict[chain_id]

            resseq = atom_entry.label_seq_id
            auth_seq_id = atom_entry.auth_seq_id
            if resseq is None:
                # The residue sequence for HETATOM is not defined in mmCIF, 
                # we need to manually increment it from the last sequence number
                if auth_seq_id != last_auth_seq:
                    het_chain_resseq += 1
                    last_auth_seq = auth_seq_id
                resseq = het_chain_resseq

            if resseq not in res_dict:
                res_dict[resseq] = self._create_residue_dict_entry(atom_entry, resseq)
            atom = self.create_atom_from_namedtuple(atom_entry)
            res_dict[resseq]['atom_list'].append(atom)

        return model_dict

    def create_assembly_dict(self):
        """Return a dictionary of chains grouped by mmcif assembly ids under
        "pdbx_struct_assembly_gen". If no assembly info available, None is
        returned.
        """
        self.structure_id = self.cifdict.get('data')
        if self.structure_id.startswith('AF-'):
            # AF-xxxxx is AlphaFold Structure, which does not have assembly
            return

        if 'pdbx_struct_assembly_gen' not in self.cifdict:
            warnings.warn(
                f"No structure assembly info in {self.structure_id}"
            )
            return

        assemblies = dict()
        entries = self.cifdict.create_namedtuples('pdbx_struct_assembly_gen')
        if self.use_bio_assembly:
            entries = entries[:1]
        for entry in entries:
            chain_id_list = entry.asym_id_list.split(',')
            assemblies[entry.assembly_id] = chain_id_list

        return assemblies

    def _cif_find_symmetry_info(self):
        symmetry_info = self.cifdict.get('pdbx_struct_oper_list')
        if symmetry_info is None or self.assembly_dict is None:
            return
        operation_names = symmetry_info['type']
        matrices = np.empty((len(operation_names), 3, 3))
        vectors = np.empty((len(operation_names), 3))
        for key, val in symmetry_info.items():
            if key.startswith('matrix'):
                idx_j, idx_k = int(key[-2])-1, int(key[-1])-1
                for idx_i, x in enumerate(val):
                    matrices[idx_i, idx_j, idx_k] = x
            if key.startswith('vector'):
                idx_j = int(key[-1])-1
                for idx_i, x in enumerate(val):
                    vectors[idx_i, idx_j] = x

        operation_dict = {}
        for i, (operation, matrix, vector) in enumerate(
            zip(operation_names, matrices, vectors)
        ):
            operation_dict[i] = {
                'type': operation,
                'matrix': matrix,
                'vector': vector
            }

        return operation_dict

    def _oper_str_to_list(self, ops):
        ops = ops.lstrip('(').rstrip(')')
        if ',' in ops:
            return [int(op)-1 for op in ops.split(',')]
        elif '-' in ops:
            st, end = ops.split('-')
            return list(range(int(st)-1, int(end)))
        raise ValueError(f'Unknown Type of operation number format {ops}')

    def _find_first_assembly_ops(self):
        oper_details = self.cifdict.create_namedtuples('pdbx_struct_assembly_gen')
        for oper_detail in oper_details:
            if oper_detail.assembly_id != 1:
                continue
            op_val = oper_detail.oper_expression
            if isinstance(op_val, str):
                first_assem_ops = self._oper_str_to_list(op_val)
            else:
                first_assem_ops = [int(op_val)-1]
        return first_assem_ops

    def _execute_symmetry_operations(self, model):
        if not self.symmetry_ops:
            return
        reference_model = deepcopy(model)
        reference_model.detach_parent()
        idx_translations = [letters_to_index(chain.id) for chain in model]
        last_idx = max(idx_translations)
        for op_id in self._find_first_assembly_ops():
            if op_id not in self.symmetry_ops:
                raise ValueError(
                    f'Operation id {op_id} not found in listed symmetry ops'
                )
            operation = self.symmetry_ops[op_id]
            if operation['type'] == 'identity operation':
                continue
            warnings.warn(
                f"{operation['type'].upper()} performed as specified in mmCIF file."
            )
            matrix, vector = operation['matrix'], operation['vector']
            copy_model = deepcopy(reference_model)
            copy_coords = get_coords(copy_model)
            new_coords = copy_coords @ matrix + vector
            for i, atom in enumerate(copy_model.get_atoms()):
                atom.coord = new_coords[i]
            for chain in copy_model:
                last_idx += 1
                chain.detach_parent()
                chain.id = index_to_letters(last_idx)
            for chain in copy_model:
                model.add(chain)
        
        model.reset_atom_serial_numbers()

    def _build_structure(self, structure_id):
        """build the structure with structure builder object and mmcif dict"""
        sb = self._structure_builder
        self.model_template = self.create_model_template()
        
        all_anisou = self.cifdict.create_namedtuples('atom_site_anisotrop')
        if structure_id is None:
            structure_id = self.cifdict['data']
        sb.init_structure(structure_id)
        sb.init_seg(" ")
        if 'cell' in self.cifdict:
            cell_info = self.cifdict.create_namedtuples('cell')[0]
            sb.structure.cell_info = cell_info._asdict()
        
        if self.assembly_dict is None:
            selected_chains = None
        else:
            sb.structure.assemblies = self.assembly_dict
            selected_chains = []
            for chain_list in self.assembly_dict.values():
                selected_chains.extend(chain_list)

        atom_site = self.create_atom_site_entry_dict()
        ##TODO: refactor these nested for loops
        for model_id, entity_dict in atom_site.items():
            model = Model(model_id)
            sb.structure.add(model)
            sb.model = model
            for entity_id, chain_dict in entity_dict.items():
                for chain_id, res_dict in chain_dict.items():
                    if (
                        selected_chains is not None
                    ) and (
                        chain_id not in selected_chains
                    ):
                        self.skipped_chain_id.append(chain_id)
                        continue
                    chain = self.model_template[entity_id].copy()
                    if isinstance(chain, Solvent):
                        if not self.include_solvent:
                            self.skipped_chain_id.append(chain_id)
                            continue
                        chain.source = 'PDB'
                    sb.model.add(chain)
                    sb.chain = chain
                    # This is the mmCIF label chain id
                    sb.chain.id = chain_id
                    for resseq, res_info in res_dict.items():
                        resname = res_info["resname"]
                        res_id = res_info["res_id"]
                        author_seq_id = res_info["author_seq_id"]
                        atoms = res_info["atom_list"]
                        sb.init_residue(resname, *res_id, author_seq_id=author_seq_id)
                        for atom in atoms:
                            sb.add_atom(atom, sb.residue)
                            if atom.orig_serial_number <= len(all_anisou):
                                anisou = all_anisou[atom.orig_serial_number-1]
                                u = (
                                    anisou.U11, anisou.U12, anisou.U13,
                                    anisou.U22, anisou.U23, anisou.U33,
                                )
                                atom.set_anisou(np.array(u, "f"))
            if self.use_bio_assembly:
                self._execute_symmetry_operations(sb.model)

        for model in sb.structure:
            for chain in model:
                if isinstance(chain, PolymerChain):
                    chain.reset_disordered_residues()
                    chain.sort_residues()
                elif isinstance(chain, Heterogens):
                    chain.update()
        sb.structure.set_pdb_id(structure_id)
        self.add_cell_and_symmetry_info()
        self.add_connect_record()
        
    def add_cell_and_symmetry_info(self):
        cell = self.cifdict.create_namedtuples('cell', single_value=True)
        symmetry = self.cifdict.create_namedtuples('symmetry', single_value=True)

        if not (
            cell and symmetry and hasattr(symmetry, "space_group_name_H_M")
        ):
            return
        
        a = float(cell.length_a)
        b = float(cell.length_b)
        c = float(cell.length_c)
        alpha = float(cell.angle_alpha)
        beta = float(cell.angle_beta)
        gamma = float(cell.angle_gamma)
        cell_data = np.array((a, b, c, alpha, beta, gamma), "f")
        spacegroup = symmetry.space_group_name_H_M #Hermann-Mauguin space-group symbol
        self._structure_builder.set_symmetry(spacegroup, cell_data)

    def add_connect_record(self):
        if 'struct_conn' not in self.cifdict:
            return
        struct_conn = self.cifdict.create_namedtuples('struct_conn')
        label_dict = {
            'chain': 'ptnr{}_label_asym_id',
            'resname': 'ptnr{}_label_comp_id',
            'resseq': 'ptnr{}_label_seq_id',
            'atom_id': 'ptnr{}_label_atom_id',
            'altloc': 'pdbx_ptnr{}_label_alt_id'
        }

        conn_dict = {}
        for connect in struct_conn:
            conn_type = connect.conn_type_id
            if conn_type not in conn_dict:
                conn_dict[conn_type] = []
            cur_connect = []
            for i in (1,2):
                cur_label_dict = {k: v.format(i) for k, v in label_dict.items()}
                record_dict = {
                    k: getattr(connect, v) for k, v in cur_label_dict.items()
                }
                if record_dict['chain'] in self.skipped_chain_id:
                    break
                cur_connect.append(record_dict)
            else:
                conn_dict[conn_type].append(tuple(cur_connect))
        self._structure_builder.set_connect(conn_dict)


import warnings

# Allowed Elements
from Bio.Data.IUPACData import atom_weights
from crimm.StructEntities.Atom import Atom

_ATOM_FORMAT_STRING = (
    "%s%5i %-4s%c%3s %c%4i%c   %8.3f%8.3f%8.3f%s%6.2f      %4s%2s%2s\n"
)
_CHARMM_ATOM_FORMAT_STRING = (
    "%s%5i %-4s%c%4s %4i%c   %8.3f%8.3f%8.3f%s%6.2f      %4s\n"
)
_TER_FORMAT_STRING = (
    "TER   %5i      %3s %c%4i%c                                                      \n"
)

def _get_atom_line_with_parent_info(
        atom: Atom, trunc_resname=False, use_charmm_format=False, convert_water=False
    ):
    """Return the parent info of the atom (PRIVATE). Atom must have a parent residue."""
    residue = atom.parent
    resname = residue.resname
    # Convert common water residue names to HOH for better visualization for nglview
    if convert_water and resname in ('TIP3', 'TIP2', 'WAT', 'TIP3P'):
        resname = 'HOH'
    segid = residue.segid
    hetfield, resseq, icode = residue.id
    if (chain:=residue.parent) is not None:
        chain_id = chain.get_id()[0]
    else:
        chain_id = '_'
    return _get_atom_line(
        atom, hetfield, segid, atom.get_serial_number(),
        resname, resseq, icode, chain_id, trunc_resname=trunc_resname,
        use_charmm_format=use_charmm_format
    )

def _get_ter_line(atom: Atom, trunc_resname=False):
    """Return the parent info of the atom (PRIVATE). Atom must have a parent residue."""
    residue = atom.parent
    resname = residue.resname
    hetfield, resseq, icode = residue.id
    if (chain:=residue.parent) is not None:
        chain_id = chain.get_id()[0]
    else:
        # no chain info, no standard TER line
        return 'TER\n'
    
    if len(resname) > 3 and trunc_resname:
        # Truncate residue name to 3 characters so it does not mess up
        # the nglview visualization
        resname = resname[:3]

    args = (atom.get_serial_number(), resname, chain_id, resseq, icode)
    return _TER_FORMAT_STRING % args

def _get_orphan_atom_line(atom: Atom, trunc_resname=False, use_charmm_format=False):
    """Return the orphan atom line (PRIVATE). Dummy residue and chain info will be filled."""
    resname = 'DUM'
    segid = ' '
    hetfield, resseq, icode = ' ', 1, ' '
    chain_id = '_'
    return _get_atom_line(
        atom, hetfield, segid, atom.get_serial_number(),
        resname, resseq, icode, chain_id, trunc_resname=trunc_resname,
        use_charmm_format=use_charmm_format
    )

def _get_atom_line(
    atom,
    hetfield,
    segid,
    atom_number,
    resname,
    resseq,
    icode,
    chain_id,
    charge="  ",
    trunc_resname=False,
    use_charmm_format=False,
):
    """Return an ATOM PDB string (PRIVATE)."""
    if hetfield != " ":
        record_type = "HETATM"
    else:
        record_type = "ATOM  "

    if use_charmm_format:
        # CHARMM format does not have chain id and record type is always ATOM
        format_string = _CHARMM_ATOM_FORMAT_STRING
        record_type = "ATOM  "
    else:
        format_string = _ATOM_FORMAT_STRING

    if len(resname) > 3 and trunc_resname:
        # Truncate residue name to 3 characters so it does not mess up
        # the nglview visualization
        resname = resname[:3]
    # Atom properties

    # Check if the atom serial number is an integer
    # Not always the case for structures built from
    # mmCIF files.
    try:
        atom_number = int(atom_number)
    except ValueError as exc:
        raise ValueError(
            f"{atom_number!r} is not a number."
            "Atom serial numbers must be numerical"
            " If you are converting from an mmCIF"
            " structure, try using"
            " preserve_atom_numbering=False"
        ) from exc

    if atom_number > 99999:
        raise ValueError(
            f"Atom serial number ('{atom_number}') exceeds PDB format limit."
        )

    # Check if the element is valid, unknown (X), or blank
    if atom.element:
        element = atom.element.strip().upper()
        if element.capitalize() not in atom_weights and element != "X":
            raise ValueError(f"Unrecognised element {atom.element}")
        element = element.rjust(2)
    else:
        element = "  "

    # Format atom name
    # Pad if:
    #     - smaller than 4 characters
    # AND - is not C, N, O, S, H, F, P, ..., one letter elements
    # AND - first character is NOT numeric (funky hydrogen naming rules)
    name = atom.fullname.strip()
    if len(name) < 4 and name[:1].isalpha() and len(element.strip()) < 2:
        name = " " + name

    altloc = atom.altloc
    x, y, z = atom.coord

    # Write PDB format line
    bfactor = atom.bfactor
    try:
        occupancy = f"{atom.occupancy:6.2f}"
    except (TypeError, ValueError):
        if atom.occupancy is None:
            occupancy = " " * 6
            warnings.warn(
                f"Missing occupancy in atom {atom.full_id!r} written as blank"
            )
        else:
            raise ValueError(
                f"Invalid occupancy value: {atom.occupancy!r}"
            ) from None

    if use_charmm_format:
        args = (
            record_type,
            atom_number,
            name,
            altloc,
            resname,
            resseq,
            icode,
            x,
            y,
            z,
            occupancy,
            bfactor,
            segid,
        )
    else:
        args = (
            record_type,
            atom_number,
            name,
            altloc,
            resname,
            chain_id,
            resseq,
            icode,
            x,
            y,
            z,
            occupancy,
            bfactor,
            segid,
            element,
            charge,
        )
    return format_string % args

##TODO: Add support for CONECT records
def get_pdb_str(
        entity, reset_serial=True,
        include_alt=False, trunc_resname=False, 
        use_charmm_format=False, convert_water=False
    ):
    """Return the PDB string of the entity.
    Parameters
    ----------
    entity : crimm.StructEntities.Entity
        The entity to convert to PDB string.
    reset_serial : bool, optional
        Whether to reset the atom serial numbers, by default True.
    include_alt : bool, optional
        Whether to include alternate location atoms, by default False.
    trunc_resname : bool, optional
        Whether to truncate residue names to 3 characters, by default False.
    use_charmm_format : bool, optional
        Whether to use CHARMM PDB format (no chain ID), by default False.
    convert_water : bool, optional
        Whether to convert common water residue names to HOH, by default False.
    Returns
    -------
    str
        The PDB string of the entity.
    """
    if reset_serial and hasattr(entity, 'reset_atom_serial_numbers'):
        entity.reset_atom_serial_numbers(include_alt=include_alt)

    if entity.level == 'A':
        if entity.parent is None:
            return _get_orphan_atom_line(entity, trunc_resname, use_charmm_format)
        return _get_atom_line_with_parent_info(
            entity, trunc_resname, use_charmm_format, convert_water
        )

    chains = []
    if entity.level in ('C', 'R', 'A'):
        chains.append(entity)
    elif entity.level == 'M':
        chains += entity.child_list
    elif entity.level == 'S':
        chains += entity.child_list[0].child_list
    
    pdb_str = ''
    for chain in chains:
        atoms = list(chain.get_atoms(include_alt=include_alt))
        if len(atoms) == 0:
            continue
        for atom in atoms:
            pdb_str += _get_atom_line_with_parent_info(
                atom, trunc_resname, use_charmm_format, convert_water
            )
        pdb_str += _get_ter_line(atoms[-1], trunc_resname)
    pdb_str += 'END\n'
    return pdb_str